from __future__ import annotations

from hashlib import md5
import datetime
import json
from typing import List, Any, Set, Optional, Tuple, Union, Literal
from pydantic import (
    BaseModel,
    ConfigDict,
    TypeAdapter,
    field_validator,
    model_validator,
    ValidationInfo,
)

import logging

logger = logging.getLogger(__name__)

from prodsys.models import port_data, performance_data
from prodsys.models import (
    dependency_data,
    node_data,
    order_data as order_data_module,
    product_data,
    resource_data,
    sink_data,
    source_data,
    time_model_data as time_model_data_module,
)
from prodsys.models import state_data as state_data_module
from prodsys.models import processes_data as processes_data_module
from prodsys.models import sink_data as sink_data_module
from prodsys.models import source_data as source_data_module
from prodsys.models import resource_data as resource_data_module
from prodsys.models import product_data as product_data_module
from prodsys.models import port_data as queue_data_module
from prodsys.models import node_data as node_data_module
from prodsys.models import scenario_data as scenario_data_module
from prodsys.models import dependency_data as dependency_data_module
from prodsys.models import primitives_data as primitives_data_module
from prodsys.util import util
from prodsys.models.processes_data import LinkTransportProcessData

def get_production_resources(
    adapter: ProductionSystemData,
) -> List[resource_data_module.ResourceData]:
    """
    Returns a list of all machines in the adapter.
    Resources are considered production resources if they have production or capability processes.

    Args:
        adapter (ProductionSystemAdapter): ProductionSystemAdapter object

    Returns:
        List[resource_data_module.ResourceData]: List of all machines in the adapter
    """
    # Get all production process IDs (flatten tuples for compound processes)
    production_process_ids = get_possible_production_processes_IDs(adapter)
    flat_production_ids = set()
    for p_id in production_process_ids:
        if isinstance(p_id, tuple):
            flat_production_ids.update(p_id)
        else:
            flat_production_ids.add(p_id)
    
    # Return resources that have at least one production process
    return [
        resource
        for resource in adapter.resource_data
        if any(proc_id in flat_production_ids for proc_id in resource.process_ids)
    ]


def get_transport_resources(
    adapter: ProductionSystemData,
) -> List[resource_data_module.ResourceData]:
    """
    Returns a list of all transport resources in the adapter.
    Resources are considered transport resources if they have transport processes.

    Args:
        adapter (ProductionSystemAdapter): ProductionSystemAdapter object

    Returns:
        List[resource_data_module.ResourceData]: List of all transport resources in the adapter
    """
    # Get all transport process IDs
    transport_process_ids = set(get_possible_transport_processes_IDs(adapter))
    
    # Return resources that have at least one transport process
    return [
        resource
        for resource in adapter.resource_data
        if any(proc_id in transport_process_ids for proc_id in resource.process_ids)
    ]


def get_set_of_IDs(list_of_objects: List[Any]) -> Set[str]:
    """
    Returns a set of all IDs of the objects in the list, by utilizing the ID attribute of the objects.

    Args:
        list_of_objects (List[Any]): List of objects that have an ID attribute

    Returns:
        Set[str]: Set of all IDs of the objects in the list
    """
    return set([obj.ID for obj in list_of_objects])


def has_only_transport_processes(resource: resource_data_module.ResourceData, available_processes: List[processes_data_module.PROCESS_DATA_UNION]) -> bool:
    process_objects = [process for process in available_processes if process.ID in resource.process_ids]
    return all(process.type in [
        processes_data_module.ProcessTypeEnum.TransportProcesses, 
    processes_data_module.ProcessTypeEnum.LinkTransportProcesses
    ] for process in process_objects)


def get_default_queue_for_resource(
    resource: resource_data_module.ResourceData,
    adapter: ProductionSystemData,
    queue_capacity: Union[float, int] = 0.0,
) -> Optional[queue_data_module.QueueData]:
    """
    Returns default input and output queues for the given resource.

    Args:
        resource (resource_data_module.ResourceData): Resource for which the default queues should be returned
        queue_capacity (Union[float, int], optional): Capacity of the default queues. Defaults to 0.0 (infinite queue).
        adapter (Optional[ProductionSystemData], optional): Production system adapter containing process data. 
                                                           Required for accurate transport resource detection. Defaults to None.

    Returns:
        tuple[queue_data_module.QueueData, queue_data_module.QueueData]: Tuple of (input_queue, output_queue)
    """
    # Determine if this is a transport resource by checking:
    # 1. First priority: check if resource has transport processes (if adapter provided)
    # 2. Fallback: check can_move attribute
    if has_only_transport_processes(resource, adapter.process_data):
        return None
    
    # For production resources, create separate input and output queues
    queue = queue_data_module.QueueData(
            ID=resource.ID + "_default_queue",
            description="Default queue of " + resource.ID,
            capacity=queue_capacity,
            location=resource.location,
            interface_type=port_data.PortInterfaceType.INPUT_OUTPUT,
            port_type=port_data.PortType.QUEUE,
        )
    return queue

def remove_queues_from_resource(machine: resource_data_module.ResourceData):
    machine.ports = []


def remove_queues_from_resources(
    machines: List[resource_data_module.ResourceData],
):
    for machine in machines:
        remove_queues_from_resource(machine)


def remove_queues_from_source(source: source_data_module.SourceData):
    source.ports = []


def remove_queues_from_sink(sink: sink_data_module.SinkData):
    sink.ports = []


def remove_unused_queues_from_adapter(
    adapter: ProductionSystemData,
) -> ProductionSystemData:
    used_queues_ids = set(
        [queue_ID for machine in adapter.resource_data for queue_ID in (machine.ports or [])]
        + [queue_ID for source in adapter.source_data for queue_ID in (source.ports or [])]
        + [queue_ID for sink in adapter.sink_data for queue_ID in (sink.ports or [])]
        + [
            queue_ID
            for primitive in adapter.primitive_data
            for queue_ID in primitive.storages
        ]
    )
    queues_to_remove = []
    for queue in adapter.port_data:
        if queue.ID not in used_queues_ids:
            queues_to_remove.append(queue)
    for queue in queues_to_remove:
        adapter.port_data.remove(queue)
    return adapter

def has_input_port(locatable: resource_data_module.ResourceData, adapter: ProductionSystemData) -> bool:
    if not locatable.ports:
        return False
    ports = [port for port in adapter.port_data if port.ID in locatable.ports]
    return any(port.interface_type in [port_data.PortInterfaceType.INPUT, port_data.PortInterfaceType.INPUT_OUTPUT] for port in ports)

def has_output_port(locatable: resource_data_module.ResourceData, adapter: ProductionSystemData) -> bool:
    if not locatable.ports:
        return False
    ports = [port for port in adapter.port_data if port.ID in locatable.ports]
    return any(port.interface_type in [port_data.PortInterfaceType.OUTPUT, port_data.PortInterfaceType.INPUT_OUTPUT] for port in ports)

def add_default_queues_to_resources(
    adapter: ProductionSystemData, queue_capacity=0.0, reset=True
) -> ProductionSystemData:
    """
    Convenience function to add default queues to all resources in the adapter.
    Creates both input and output queues for production resources, and a single 
    input_output queue for transport resources.

    Args:
        adapter (ProductionSystemAdapter): ProductionSystemAdapter object
        queue_capacity (float, optional): Capacity of the default queues. Defaults to 0.0 (infinite queue).
        reset (bool, optional): If True, removes existing ports and creates new ones. If False, only adds ports if missing. Defaults to True.

    Returns:
        ProductionSystemAdapter: ProductionSystemAdapter object with default queues added to all resources
    """
    for resource in adapter.resource_data:
        transport_process_ids = set(get_possible_transport_processes_IDs(adapter))
        is_transport_resource = any(proc_id in transport_process_ids for proc_id in resource.process_ids)

        if (not reset and has_input_port(resource, adapter) and has_output_port(resource, adapter)) or is_transport_resource:
            continue
        remove_queues_from_resource(resource)
        remove_unused_queues_from_adapter(adapter)
        
        # Get both input and output queues, passing adapter for proper transport resource detection
        default_queue = get_default_queue_for_resource(
            resource, adapter, queue_capacity
        )
        if default_queue is not None:
            # Check if queue with this ID already exists before adding
            existing_queue_ids = {queue.ID for queue in adapter.port_data}
            if default_queue.ID not in existing_queue_ids:
                # Add queues to adapter only if it doesn't already exist
                adapter.port_data.append(default_queue)
            # Always set the resource's ports reference, even if queue already exists
            resource.ports = [default_queue.ID]
    return adapter


def get_default_queue_for_source(
    source: source_data_module.SourceData, queue_capacity=0.0
) -> queue_data_module.QueueData:
    """
    Returns a default queue for the given source.

    Args:
        source (source_data_module.SourceData): Source for which the default queue should be returned
        queue_capacity (float, optional): Capacity of the default queue. Defaults to 0.0 (infinite queue).

    Returns:
        queue_data_module.QueueData: Default queue for the given source
    """
    return queue_data_module.QueueData(
        ID=source.ID + "_default_output_queue",
        description="Default output queue of " + source.ID,
        capacity=queue_capacity,
        location=source.location,
        interface_type=port_data.PortInterfaceType.OUTPUT,
        port_type=port_data.PortType.QUEUE,
    )


def add_default_queues_to_sources(
    adapter: ProductionSystemData, queue_capacity=0.0, reset=True
) -> ProductionSystemData:
    """
    Convenience function to add default queues to all sources in the adapter.

    Args:
        adapter (ProductionSystemAdapter): ProductionSystemAdapter object
        queue_capacity (float, optional): Capacity of the default queues. Defaults to 0.0 (infinite queue).

    Returns:
        ProductionSystemAdapter: ProductionSystemAdapter object with default queues added to all sources
    """
    for source in adapter.source_data:
        if not reset and has_output_port(source, adapter):
            continue
        remove_queues_from_source(source)
        remove_unused_queues_from_adapter(adapter)

        output_queues = [get_default_queue_for_source(source, queue_capacity)]
        source.ports = list(get_set_of_IDs(output_queues))
        adapter.port_data += output_queues
    return adapter


def get_default_queue_for_sink(
    sink: sink_data_module.SinkData, queue_capacity=0.0
) -> queue_data_module.QueueData:
    """
    Returns a default queue for the given sink.

    Args:
        sink (sink_data_module.SinkData): Sink for which the default queue should be returned
        queue_capacity (float, optional): Capacity of the default queue. Defaults to 0.0 (infinite queue).

    Returns:
        queue_data_module.QueueData: Default queue for the given sink
    """
    return queue_data_module.QueueData(
        ID=sink.ID + "_default_input_queue",
        description="Default input queue of " + sink.ID,
        capacity=queue_capacity,
        location=sink.location,
        interface_type=port_data.PortInterfaceType.INPUT,
        port_type=port_data.PortType.QUEUE,
    )


def add_default_queues_to_sinks(
    adapter: ProductionSystemData, queue_capacity=0.0, reset=True
) -> ProductionSystemData:
    """
    Convenience function to add default queues to all sinks in the adapter.

    Args:
        adapter (ProductionSystemAdapter): ProductionSystemAdapter object
        queue_capacity (float, optional): Capacity of the default queues. Defaults to 0.0 (infinite queue).

    Returns:
        ProductionSystemAdapter: ProductionSystemAdapter object with default queues added to all sinks
    """
    for sink in adapter.sink_data:
        if not reset and has_input_port(sink, adapter):
            continue
        remove_queues_from_sink(sink)
        remove_unused_queues_from_adapter(adapter)

        input_queues = [get_default_queue_for_sink(sink, queue_capacity)]
        sink.ports = list(get_set_of_IDs(input_queues))
        adapter.port_data += input_queues
    return adapter


def add_default_queues_to_production_system(
    adapter: ProductionSystemData, queue_capacity=0.0, reset=True
) -> ProductionSystemData:
    """
    Convenience function to add default queues to all machines, sources and sinks in the adapter.

    Args:
        adapter (ProductionSystemAdapter): ProductionSystemAdapter object
        queue_capacity (float, optional): Capacity of the default queues. Defaults to 0.0 (infinite queue).

    Returns:
        ProductionSystemAdapter: ProductionSystemAdapter object with default queues added to all machines, sources and sinks
    """
    adapter = add_default_queues_to_resources(adapter, queue_capacity, reset)
    adapter = add_default_queues_to_sources(adapter, queue_capacity, reset)
    adapter = add_default_queues_to_sinks(adapter, queue_capacity, reset)
    return adapter


def load_json(file_path: str) -> dict:
    with open(file_path, "r", encoding="utf-8") as json_file:
        data = json.load(json_file)
    return data


class ProductionSystemData(BaseModel):
    """
    A ProductionSystemAdapter serves as a n abstract base class of a data container to represent a production system. It is based on the `prodsys.models` module, but is also compatible with the `prodsys.express` API.
    It is used as the basis for all simulation and optimization algorithms in prodsys and comes with complete data validation.
    Thereby, it is assured that the expected data is used for simulation and optimization. If the data is not valid, an error is raised with information about the reasons for invalidity.
    The adapter targets easy integration of algorithms with each other in different environments.
    Therefore, the adapter can even be used for integration of new algorithms by serving as a defined data interface.

    Attributes:
        ID: ID of the production system. Defaults to "".
        seed: Seed for the random number generator used in simulation. Defaults to 0.
        time_model_data: List of time models used by the entities in the production system. Defaults to [].
        state_data: List of states used by the resources in the production system. Defaults to [].
        process_data: List of processes required by products and provided by resources in the production system. Defaults to [].
        port_data: List of ports used by the resources, sources and sinks in the production system. Defaults to [].
        node_data: List of nodes in the production system. Defaults to [].
        resource_data: List of resources in the production system. Defaults to [].
        product_data: List of products in the production system. Defaults to [].
        sink_data: List of sinks in the production system. Defaults to [].
        source_data: List of sources in the production system. Defaults to [].
        scenario_data: Scenario data of the production system used for optimization. Defaults to None.
        dependency_data: List of dependencies in the production system. Defaults to [].
        primitive_data: List of stored primitives in the production system. Defaults to [].
        schedule: List of scheduled Events of the production system. Defaults to None.
        order_data: List of orders in the production system. Defaults to None.
        conwip_number: Number of allowed WIP (Work in Progress - number of released products) in the production system. Defaults to None.
        reference_time: Reference time of the production system. Defaults to None.
        time_unit: Time unit of the production system. Defaults to "min".
        valid_configuration: Indicates if the configuration is valid. Defaults to True.
        reconfiguration_cost: Cost of reconfiguration in a optimization scenario. Defaults to 0.
    """

    # TODO: add check, that throws an error, if items have the same ID!
    ID: str = ""
    seed: int = 0
    time_model_data: List[time_model_data_module.TIME_MODEL_DATA] = []
    state_data: List[state_data_module.STATE_DATA_UNION] = []
    process_data: List[processes_data_module.PROCESS_DATA_UNION] = []
    port_data: List[queue_data_module.QUEUE_DATA_UNION] = []
    node_data: List[node_data_module.NodeData] = []
    resource_data: List[resource_data_module.ResourceData] = []
    product_data: List[product_data_module.ProductData] = []
    sink_data: List[sink_data_module.SinkData] = []
    source_data: List[source_data_module.SOURCE_DATA_UNION] = []
    scenario_data: Optional[scenario_data_module.ScenarioData] = None
    dependency_data: Optional[List[dependency_data_module.DEPENDENCY_TYPES]] = []
    primitive_data: Optional[List[primitives_data_module.StoredPrimitive]] = []
    schedule: Optional[List[performance_data.Event]] = None
    order_data: Optional[List[order_data_module.OrderData]] = None
    conwip_number: Optional[int] = None
    reference_time: Optional[datetime.datetime] = None
    time_unit: Literal["s", "min", "h", "d"] = "min"

    valid_configuration: bool = True
    reconfiguration_cost: float = 0

    model_config = ConfigDict(
        validate_assignment=True,
        revalidate_instances='always',
        json_schema_extra={
            "examples": [
                # TODO: update data here
                {
                    "ID": "Example Adapter",
                    "valid_configuration": True,
                    "reconfiguration_cost": 0,
                    "seed": 24,
                    "time_model_data": [
                        {
                            "ID": "function_time_model_1",
                            "description": "normal distribution time model with 20 minutes",
                            "distribution_function": "normal",
                            "location": 14.3,
                            "scale": 5.0,
                            "batch_size": 100,
                        },
                        {
                            "ID": "function_time_model_2",
                            "description": "constant distribution time model with 10 minutes",
                            "distribution_function": "constant",
                            "location": 15.0,
                            "scale": 0.0,
                            "batch_size": 100,
                        },
                        {
                            "ID": "function_time_model_3",
                            "description": "normal distribution time model with 20 minutes",
                            "distribution_function": "normal",
                            "location": 20.0,
                            "scale": 5.0,
                            "batch_size": 100,
                        },
                        {
                            "ID": "function_time_model_4",
                            "description": "exponential distribution time model with 100 minutes",
                            "distribution_function": "exponential",
                            "location": 52.0,
                            "scale": 0.0,
                            "batch_size": 100,
                        },
                        {
                            "ID": "function_time_model_5",
                            "description": "exponential distribution time model with 150 minutes",
                            "distribution_function": "exponential",
                            "location": 150.0,
                            "scale": 0.0,
                            "batch_size": 100,
                        },
                        {
                            "ID": "sequential_time_model_1",
                            "description": "Sequential time model",
                            "sequence": [25.0, 13.0, 15.0, 16.0, 17.0, 20.0, 21.0],
                        },
                        {
                            "ID": "manhattan_time_model_1",
                            "description": "manhattan time model with speed 180 m/min = 3 m/s",
                            "speed": 30.0,
                            "reaction_time": 0.15,
                        },
                        {
                            "ID": "function_time_model_7",
                            "description": "exponential distribution time model with 300 minutes",
                            "distribution_function": "exponential",
                            "location": 300.0,
                            "scale": 0.0,
                            "batch_size": 100,
                        },
                        {
                            "ID": "function_time_model_8",
                            "description": "normal distribution time model with 15 minutes",
                            "distribution_function": "normal",
                            "location": 15.0,
                            "scale": 3.0,
                            "batch_size": 100,
                        },
                    ],
                    "state_data": [
                        {
                            "ID": "Breakdownstate_1",
                            "description": "Breakdown state machine 1",
                            "time_model_id": "function_time_model_5",
                            "type": "BreakDownState",
                            "repair_time_model_id": "function_time_model_8",
                        },
                        {
                            "ID": "Breakdownstate_2",
                            "description": "Breakdown state machine 2",
                            "time_model_id": "function_time_model_5",
                            "type": "BreakDownState",
                            "repair_time_model_id": "function_time_model_8",
                        },
                        {
                            "ID": "Setup_State_1",
                            "description": "Setup state machine 1",
                            "time_model_id": "function_time_model_2",
                            "type": "SetupState",
                            "origin_setup": "P1",
                            "target_setup": "P2",
                        },
                        {
                            "ID": "Setup_State_2",
                            "description": "Setup state machine 2",
                            "time_model_id": "function_time_model_2",
                            "type": "SetupState",
                            "origin_setup": "P2",
                            "target_setup": "P1",
                        },
                        {
                            "ID": "Setup_State_3",
                            "description": "Setup state machine 3",
                            "time_model_id": "function_time_model_2",
                            "type": "SetupState",
                            "origin_setup": "P1",
                            "target_setup": "P3",
                        },
                        {
                            "ID": "Setup_State_4",
                            "description": "Setup state machine 3",
                            "time_model_id": "function_time_model_3",
                            "type": "SetupState",
                            "origin_setup": "P3",
                            "target_setup": "P1",
                        },
                        {
                            "ID": "ProcessBreakdownState_1",
                            "description": "Breakdown state process 1",
                            "time_model_id": "function_time_model_7",
                            "type": "ProcessBreakDownState",
                            "repair_time_model_id": "function_time_model_8",
                            "process_id": "P1",
                        },
                    ],
                    "process_data": [
                        {
                            "ID": "P1",
                            "description": "Process 1",
                            "time_model_id": "function_time_model_1",
                            "type": "ProductionProcesses",
                        },
                        {
                            "ID": "P2",
                            "description": "Process 2",
                            "time_model_id": "function_time_model_2",
                            "type": "ProductionProcesses",
                        },
                        {
                            "ID": "P3",
                            "description": "Process 3",
                            "time_model_id": "function_time_model_3",
                            "type": "ProductionProcesses",
                        },
                        {
                            "ID": "TP1",
                            "description": "Transport Process 1",
                            "time_model_id": "manhattan_time_model_1",
                            "loading_time_model_id": "function_time_model_2",
                            "unloading_time_model_id": "function_time_model_3",
                            "type": "TransportProcesses",
                        },
                    ],
                    "port_data": [
                        {
                            "ID": "IQ1",
                            "description": "Input-queue 1 for R1",
                            "capacity": 10,
                        },
                        {
                            "ID": "OQ1",
                            "description": "Output-queue 1 for R1",
                            "capacity": 10,
                        },
                        {
                            "ID": "OQ2",
                            "description": "Output-queue 2 for R2",
                            "capacity": 10,
                        },
                        {
                            "ID": "IQ2",
                            "description": "Input-queue 2 for R3",
                            "capacity": 10,
                        },
                        {
                            "ID": "OQ3",
                            "description": "Output-queue 3 for R3",
                            "capacity": 10,
                        },
                        {
                            "ID": "SourceQueue",
                            "description": "Output-Queue for all sources",
                            "capacity": 0,
                        },
                        {
                            "ID": "SinkQueue",
                            "description": "Input-Queue for all sinks",
                            "capacity": 0,
                        },
                        {
                            "ID": "IQ9",
                            "description": "Input-queue 1 for R2",
                            "capacity": 10,
                        },
                    ],
                    "resource_data": [
                        {
                            "ID": "R1",
                            "description": "Resource 1",
                            "capacity": 2,
                            "input_location": [10.0, 10.0],
                            "output_location": [12.0, 10.0],
                            "controller": "PipelineController",
                            "control_policy": "FIFO",
                            "process_ids": ["P1", "P2"],
                            "process_capacities": [2, 1],
                            "state_ids": [
                                "Breakdownstate_1",
                                "Setup_State_1",
                                "Setup_State_2",
                                "ProcessBreakdownState_1",
                            ],
                            "input_queues": ["IQ1"],
                            "output_queues": ["OQ1"],
                        },
                        {
                            "ID": "R2",
                            "description": "Resource 2",
                            "capacity": 1,
                            "input_location": [20.0, 10.0],
                            "output_location": [22.0, 10.0],
                            "controller": "PipelineController",
                            "control_policy": "FIFO",
                            "process_ids": ["P2", "P3"],
                            "process_capacities": None,
                            "state_ids": ["Breakdownstate_2"],
                            "input_queues": ["IQ9"],
                            "output_queues": ["OQ2"],
                        },
                        {
                            "ID": "R3",
                            "description": "Resource 3",
                            "capacity": 2,
                            "input_location": [20.0, 20.0],
                            "output_location": [22.0, 22.0],
                            "controller": "PipelineController",
                            "control_policy": "FIFO",
                            "process_ids": ["P1", "P3"],
                            "process_capacities": [1, 2],
                            "state_ids": [
                                "Breakdownstate_1",
                                "Breakdownstate_2",
                                "Setup_State_3",
                                "Setup_State_4",
                            ],
                            "input_queues": ["IQ2"],
                            "output_queues": ["OQ3"],
                        },
                        {
                            "ID": "R4",
                            "description": "Resource 3",
                            "capacity": 2,
                            "input_location": [10.0, 20.0],
                            "output_location": [12.0, 20.0],
                            "controller": "PipelineController",
                            "control_policy": "FIFO",
                            "process_ids": ["P1", "P3"],
                            "process_capacities": [2, 2],
                            "state_ids": [
                                "Breakdownstate_1",
                                "Setup_State_3",
                                "Setup_State_4",
                            ],
                            "input_queues": ["IQ2"],
                            "output_queues": ["OQ3"],
                        },
                        {
                            "ID": "TR1",
                            "description": "Transport Resource 1",
                            "capacity": 1,
                            "location": [15.0, 15.0],
                            "controller": "TransportController",
                            "control_policy": "FIFO",
                            "process_ids": ["TP1"],
                            "process_capacities": None,
                            "state_ids": ["Breakdownstate_1"],
                        },
                        {
                            "ID": "TR2",
                            "description": "Transport Resource 2",
                            "capacity": 1,
                            "location": [15.0, 20.0],
                            "controller": "TransportController",
                            "control_policy": "SPT_transport",
                            "process_ids": ["TP1"],
                            "process_capacities": None,
                            "state_ids": ["Breakdownstate_1"],
                        },
                    ],
                    "product_data": [
                        {
                            "ID": "Product_1",
                            "description": "Product 1",
                            "product_type": "Product_1",
                            "processes": ["P1", "P2", "P3"],
                            "transport_process": "TP1",
                        },
                        {
                            "ID": "Product_2",
                            "description": "Product 2",
                            "product_type": "Product_2",
                            "processes": ["P1", "P2", "P3", "P1"],
                            "transport_process": "TP1",
                        },
                        {
                            "ID": "Product_3",
                            "description": "Product 3",
                            "product_type": "Product_3",
                            "processes": {"P1": ["P2", "P3"], "P2": [], "P3": []},
                            "transport_process": "TP1",
                        },
                    ],
                    "sink_data": [
                        {
                            "ID": "SK1",
                            "description": "Sink 1",
                            "input_location": [50.0, 50.0],
                            "product_type": "Product_1",
                            "input_queues": ["SinkQueue"],
                        },
                        {
                            "ID": "SK2",
                            "description": "Sink 2",
                            "input_location": [55.0, 50.0],
                            "product_type": "Product_2",
                            "input_queues": ["SinkQueue"],
                        },
                        {
                            "ID": "SK3",
                            "description": "Sink 3",
                            "input_location": [45.0, 50.0],
                            "product_type": "Product_3",
                            "input_queues": ["SinkQueue"],
                        },
                    ],
                    "source_data": [
                        {
                            "ID": "S1",
                            "description": "Source 1",
                            "output_location": [0.0, 0.0],
                            "product_type": "Product_1",
                            "time_model_id": "function_time_model_4",
                            "router": "SimpleRouter",
                            "routing_heuristic": "shortest_queue",
                            "output_queues": ["SourceQueue"],
                        },
                        {
                            "ID": "S2",
                            "description": "Source 2",
                            "output_location": [30.0, 30.0],
                            "product_type": "Product_2",
                            "time_model_id": "function_time_model_4",
                            "router": "SimpleRouter",
                            "routing_heuristic": "shortest_queue",
                            "output_queues": ["SourceQueue"],
                        },
                        {
                            "ID": "S3",
                            "description": "Source 3",
                            "output_location": [40.0, 30.0],
                            "product_type": "Product_3",
                            "time_model_id": "function_time_model_4",
                            "router": "SimpleRouter",
                            "routing_heuristic": "shortest_queue",
                            "output_queues": ["SourceQueue"],
                        },
                    ],
                    "scenario_data": None,
                }
            ]
        },
    )

    @classmethod
    def read(cls, filepath: str) -> ProductionSystemData:
        """
        Reads a JSON file and returns a ProductionSystemData object.

        Args:
            filepath (str): Path to the JSON file

        Returns:
            ProductionSystemData: ProductionSystemData object
        """
        data = load_json(filepath)
        return cls.model_validate(data)

    def write(self, filepath: str):
        """
        Writes the ProductionSystemData object to a JSON file.

        Args:
            filepath (str): Path to the JSON file
        """
        with open(filepath, "w", encoding="utf-8") as json_file:
            json_file.write(self.model_dump_json(indent=4))

    def revalidate(self) -> 'ProductionSystemData':
        """
        Explicitly revalidates the entire ProductionSystemData model.
        
        This is useful after making in-place modifications to lists (e.g., appending items)
        to ensure all validators run, including the duplicate ID check.
        
        Example:
            ```python
            system = ProductionSystemData(...)
            system.node_data.append(new_node)  # In-place modification
            system.revalidate()  # Trigger validation
            ```
        
        Returns:
            ProductionSystemData: Self, after validation
            
        Raises:
            ValueError: If validation fails (e.g., duplicate IDs detected)
        """
        # Trigger validation by re-parsing the model
        validated = self.model_validate(self.model_dump())
        # Update self with validated data
        for key, value in validated.__dict__.items():
            setattr(self, key, value)
        return self

    def hash(self) -> str:
        """
        Generates a hash of the adapter based on the hash of all contained entities. Only information describing the physical structure and functionality of the production system is considered. Can be used to compare two production systems of adapters for functional equality.

        Returns:
            str: Hash of the adapter
        """
        return md5(
            (
                "".join(
                    [
                        *sorted(
                            [time_model.hash() for time_model in self.time_model_data]
                        ),
                        *sorted([state.hash(self) for state in self.state_data]),
                        *sorted([process.hash(self) for process in self.process_data]),
                        *sorted([res.hash(self) for res in self.resource_data]),
                        *sorted([queue.hash() for queue in self.port_data]),
                        *sorted([node.hash() for node in self.node_data]),
                        *sorted([product.hash(self) for product in self.product_data]),
                        *sorted([sink.hash(self) for sink in self.sink_data]),
                        *sorted([source.hash(self) for source in self.source_data]),
                        *sorted(
                            [
                                dependency.hash(self)
                                for dependency in self.dependency_data
                            ]
                        ),
                        *sorted(
                            [primitive.hash(self) for primitive in self.primitive_data]
                        ),
                    ]
                )
            ).encode("utf-8")
        ).hexdigest()

    @model_validator(mode='after')
    def check_no_duplicate_ids(self) -> 'ProductionSystemData':
        """
        Validates that no duplicate IDs exist across all data types in the production system.
        This is critical for proper object resolution, especially for link transport processes.
        
        This validator runs automatically:
        - On initialization of a new ProductionSystemData instance
        - When entire fields are reassigned (e.g., `system.node_data = new_list`)
        
        For in-place mutations (e.g., `system.node_data.append(item)`), call `revalidate()` 
        or `model_validate()` explicitly to trigger validation.
        
        Raises:
            ValueError: If duplicate IDs are found with details about which IDs and where.
        """
        id_registry = {}  # Maps ID to list of (field_name, object_type) tuples
        
        # Define all fields to check
        fields_to_check = [
            ('time_model_data', self.time_model_data, 'TimeModel'),
            ('state_data', self.state_data, 'State'),
            ('process_data', self.process_data, 'Process'),
            ('port_data', self.port_data, 'Port/Queue'),
            ('node_data', self.node_data, 'Node'),
            ('resource_data', self.resource_data, 'Resource'),
            ('product_data', self.product_data, 'Product'),
            ('sink_data', self.sink_data, 'Sink'),
            ('source_data', self.source_data, 'Source'),
            ('dependency_data', self.dependency_data or [], 'Dependency'),
            ('primitive_data', self.primitive_data or [], 'Primitive'),
        ]
        
        # Collect all IDs and their locations
        for field_name, data_list, object_type in fields_to_check:
            for item in data_list:
                if hasattr(item, 'ID'):
                    item_id = item.ID
                    if item_id not in id_registry:
                        id_registry[item_id] = []
                    id_registry[item_id].append((field_name, object_type))
        
        # Find duplicates
        duplicates = {id_: locations for id_, locations in id_registry.items() if len(locations) > 1}
        
        if duplicates:
            error_messages = []
            for duplicate_id, locations in duplicates.items():
                location_strs = [f"{obj_type} (in {field})" for field, obj_type in locations]
                error_messages.append(
                    f"  - ID '{duplicate_id}' is used in: {', '.join(location_strs)}"
                )
            
            raise ValueError(
                "\n\nDuplicate IDs detected in ProductionSystemData!\n"
                "Each ID must be unique across ALL data types (nodes, sources, sinks, resources, ports, etc.).\n"
                "This is required for proper object resolution, especially for link transport processes.\n\n"
                "Duplicate IDs found:\n" + "\n".join(error_messages) + "\n\n"
                "Resolution: Please rename the conflicting objects to have unique IDs.\n"
                "Example: If both a Node and Source have ID 'S1', rename them to 'Node_S1' and 'Source_S1'.\n"
            )
        
        return self

    @field_validator("state_data")
    def check_states(
        cls, states: List[state_data_module.STATE_DATA_UNION], info: ValidationInfo
    ):
        values = info.data
        for state in states:
            time_models = get_set_of_IDs(values["time_model_data"])
            if state.time_model_id not in time_models:
                raise ValueError(
                    f"The time model {state.time_model_id} of state {state.ID} is not a valid time model of {time_models}."
                )
        return states

    @field_validator("process_data")
    def check_processes(
        cls,
        processes: List[processes_data_module.PROCESS_DATA_UNION],
        info: ValidationInfo,
    ):
        values = info.data
        for process in processes:
            if isinstance(
                process, processes_data_module.CompoundProcessData
            ) or isinstance(
                process, processes_data_module.RequiredCapabilityProcessData
            ) or isinstance(
                process, processes_data_module.ProcessModelData
            ):
                continue
            time_models = get_set_of_IDs(values["time_model_data"])
            if process.time_model_id not in time_models:
                raise ValueError(
                    f"The time model {process.time_model_id} of process {process.ID} is not a valid time model of {time_models}."
                )
        return processes

    @field_validator("resource_data")
    def check_resources(
        cls,
        resources: List[resource_data_module.ResourceData],
        info: ValidationInfo,
    ):
        values = info.data
        for resource in resources:
            processes = get_set_of_IDs(values["process_data"])
            for process in resource.process_ids:
                if process not in processes:
                    raise ValueError(
                        f"The process {process} of resource {resource.ID} is not a valid process of {processes}."
                    )
            states = get_set_of_IDs(values["state_data"])
            for state in resource.state_ids:
                if state not in states:
                    raise ValueError(
                        f"The state {state} of resource {resource.ID} is not a valid state of {states}."
                    )
            
            # Check if resource only has transport processes
            # If it does, no queues are required

            available_processes = values["process_data"]
            available_processes = TypeAdapter(List[processes_data_module.PROCESS_DATA_UNION]).validate_python(available_processes)

            only_transport_processes = has_only_transport_processes(resource, available_processes)

            
            # If resource has no ports and doesn't only have transport processes, add default queues
            if not resource.ports and not only_transport_processes:
                queue = queue_data_module.QueueData(
                    ID=resource.ID + "_default_queue",
                    description="Default queue of " + resource.ID,
                    capacity=0.0,
                    location=resource.location,
                    interface_type=port_data.PortInterfaceType.INPUT_OUTPUT,
                    port_type=port_data.PortType.QUEUE,
                )
                if "port_data" not in values:
                    values["port_data"] = []
                values["port_data"].append(queue)
                resource.ports = [queue.ID]
                
            port_data_ids = get_set_of_IDs(values["port_data"])
            
            # Validate ports if they exist (skip validation for transport-only resources without ports)
            if resource.ports:
                for port in resource.ports:
                    if port not in port_data_ids:
                        raise ValueError(
                            f"The port {port} of resource {resource.ID} is not a valid port of {port_data_ids}."
                        )
                resource_ports: list[port_data.QueueData] = [port for port in values["port_data"] if port.ID in resource.ports]
                if not any(
                    port.interface_type
                    in [
                        queue_data_module.PortInterfaceType.OUTPUT,
                        queue_data_module.PortInterfaceType.INPUT_OUTPUT,
                    ]
                    for port in resource_ports
                ):
                    raise ValueError(
                        f"The resource {resource.ID} has no output port. Resources must have at least one output port."
                    )
                if not any(port.interface_type in [queue_data_module.PortInterfaceType.INPUT, queue_data_module.PortInterfaceType.INPUT_OUTPUT] for port in resource_ports):
                    raise ValueError(
                        f"The resource {resource.ID} has no input port. Resources must have at least one input port."
                    )
        return resources

    @field_validator("product_data")
    def check_products(
        cls, products: List[product_data_module.ProductData], info: ValidationInfo
    ):
        values = info.data
        for product in products:
            all_processes = get_set_of_IDs(values["process_data"])
            if product.transport_process not in all_processes:
                raise ValueError(
                    f"The transport process {product.transport_process} of product {product.ID} is not a valid process of {all_processes}."
                )
            required_processes = set()
            if isinstance(product.processes, list) and isinstance(
                product.processes[0], str
            ):
                required_processes = set(product.processes)
            elif isinstance(product.processes, list) and isinstance(
                product.processes[0], list
            ):
                required_processes = set(util.flatten(product.processes))
            elif isinstance(product.processes, dict):
                required_processes = set(product.processes.keys())
            if required_processes - all_processes != set():
                raise ValueError(
                    f"The processes {required_processes - all_processes} of product {product.ID} are not a valid processes of {all_processes}."
                )

        return products

    @field_validator("sink_data")
    def check_sinks(cls, sinks: List[sink_data_module.SinkData], info: ValidationInfo):
        values = info.data
        for sink in sinks:
            try:
                product_types = set([product.type for product in values["product_data"]])
            except KeyError:
                raise ValueError("Product data is missing or faulty.")
            if sink.product_type not in product_types:
                raise ValueError(
                    f"The product type {sink.product_type} of sink {sink.ID} is not a valid product type of {product_types}."
                )
            if not sink.ports:
                input_queue = get_default_queue_for_sink(sink)
                sink.ports = list(get_set_of_IDs([input_queue]))
                values["port_data"] += [input_queue]
                continue
            ports = get_set_of_IDs(values["port_data"])
            for p in sink.ports:
                if p not in ports:
                    raise ValueError(
                        f"The port {p} of sink {sink.ID} is not a valid port of {ports}."
                    )
                for port in values["port_data"]:
                    if port.ID == p:
                        if port.capacity != 0:
                            logger.warning(
                                f"The capacity of the port {port.ID} of sink {sink.ID} is limited. This might lead to unexpected behavior so it was changed to infinity."
                            )
                            port.capacity = 0
            sink_ports : list[queue_data_module.QueueData] = [port for port in values["port_data"] if port.ID in sink.ports]
            if not any(port.interface_type in [queue_data_module.PortInterfaceType.INPUT, queue_data_module.PortInterfaceType.INPUT_OUTPUT] for port in sink_ports):
                raise ValueError(
                    f"The sink {sink.ID} has no input port. Sinks must have at least one input port."
                )
        return sinks

    @field_validator("source_data")
    def check_sources(
        cls, sources: List[source_data_module.SOURCE_DATA_UNION], info: ValidationInfo
    ):
        values = info.data
        from prodsys.models.source_data import OrderSourceData
        for source in sources:
            # OrderSourceData doesn't have time_model_id or product_type
            if isinstance(source, OrderSourceData):
                # Validate order_ids exist in order_data
                if values.get("order_data"):
                    order_ids = set([order.ID for order in values["order_data"]])
                    for order_id in source.order_ids:
                        if order_id not in order_ids:
                            raise ValueError(
                                f"The order {order_id} of order source {source.ID} is not a valid order ID of {order_ids}."
                            )
            else:
                # Regular SourceData validation
                time_models = get_set_of_IDs(values["time_model_data"])
                if source.time_model_id not in time_models:
                    raise ValueError(
                        f"The time model {source.time_model_id} of source {source.ID} is not a valid time model of {time_models}."
                    )
                try:
                    product_types = set([product.type for product in values["product_data"]])
                except KeyError:
                    raise ValueError("Product data is missing or faulty.")
                if source.product_type not in product_types:
                    raise ValueError(
                        f"The product type {source.product_type} of source {source.ID} is not a valid product type of {product_types}."
                    )
            if not source.ports:
                output_queue = get_default_queue_for_source(source)
                source.ports = list(get_set_of_IDs([output_queue]))
                values["port_data"] += [output_queue]
                continue
            queues = get_set_of_IDs(values["port_data"])
            for q in source.ports:
                if q not in queues:
                    raise ValueError(
                        f"The queue {q} of source {source.ID} is not a valid queue of {queues}."
                    )
            source_ports: list[queue_data_module.QueueData] = [port for port in values["port_data"] if port.ID in source.ports]
            if not any(
                port.interface_type
                in [
                    queue_data_module.PortInterfaceType.OUTPUT,
                    queue_data_module.PortInterfaceType.INPUT_OUTPUT,
                ]
                for port in source_ports
            ):
                raise ValueError(
                    f"The source {source.ID} has no output port. Sources must have at least one output port."
                )
        return sources

    @field_validator("schedule")
    def check_schedule(
        cls, schedule: Optional[List[performance_data.Event]], info: ValidationInfo
    ):
        if schedule is None:
            return schedule
        event_resources_ids = set()
        event_process_ids = set()
        event_product_ids = set()
        schedule_to_consider = []
        for event in schedule:
            if not isinstance(event, performance_data.Event):
                raise ValueError(
                    f"The event {event} is not a valid event of {schedule}."
                )
            event_resources_ids.add(event.resource)
            product_id = "_".join(event.product.split("_")[:-1])
            event_product_ids.add(product_id)
            if event.process:
                event_process_ids.add(event.process)
            if not event.activity == "start state":
                continue
            schedule_to_consider.append(event)

        values = info.data
        resources_ids = get_set_of_IDs(values["resource_data"])
        processes_ids = get_set_of_IDs(values["process_data"])
        products_ids = get_set_of_IDs(values["product_data"])

        if event_resources_ids - resources_ids != set():
            raise ValueError(
                f"The resources {event_resources_ids - resources_ids} of the schedule are not valid resources of {resources_ids}."
            )
        if event_process_ids - processes_ids != set():
            raise ValueError(
                f"The processes {event_process_ids - processes_ids} of the schedule are not valid processes of {processes_ids}."
            )
        
        if event_product_ids - products_ids != set():
            raise ValueError(
                f"The products {event_product_ids - products_ids} of the schedule are not valid products of {products_ids}."
            )

        return schedule_to_consider

    def read_scenario(self, scenario_file_path: str):
        scenario_data = json.load(open(scenario_file_path))
        self.scenario_data = scenario_data_module.ScenarioData.model_validate_json(scenario_data)

    def validate_proceses_available(self):
        required_processes = set(
            util.flatten([product.processes for product in self.product_data])
        )
        available_processes = set()
        for resource in self.resource_data:
            for process in resource.process_ids:
                available_processes.add(process)
        if required_processes > available_processes:
            raise ValueError(
                f"The processes {required_processes - available_processes} are not available."
            )

    def validate_configuration(self):
        """
        Checks if the configuration is physically valid, i.e. if all resources are positioned at different locations and if all required processes are available.

        Raises:
            ValueError: If multiple objects are positioned at the same location.
            ValueError: If not all required process are available.
            ValueError: If not all links are available for LinkTransportProcesses.
            ValueError: If ports are missing locations (needed for transport routing).
        """
        assert_no_redundant_locations(self)
        assert_required_processes_in_resources_available(self)
        assert_all_links_available(self)
        assert_ports_have_locations(self)


def remove_duplicate_locations(input_list: List[List[float]]) -> List[List[float]]:
    return [list(x) for x in set(tuple(x) for x in input_list)]


def get_location_of_locatable(
    adapter: ProductionSystemData,
    locatable: Union[
        resource_data_module.ResourceData,
        source_data_module.SourceData,
        sink_data_module.SinkData,
        port_data.StoreData,
    ],
) -> List[List[float]]:
    locations = [locatable.location]
    if hasattr(locatable, "ports") and locatable.ports:
        for port_ID in locatable.ports:
            port = next((portx for portx in adapter.port_data if portx.ID == port_ID), None)
            if port.location != locatable.location:
                result = get_location_of_locatable(adapter, port)
                if isinstance(result, list):
                    locations.extend(result)
                else:
                    locations.append(result)
    return locations


def assert_no_redundant_locations(adapter: ProductionSystemData):
    """
    Asserts that no multiple objects are positioned at the same location.

    Args:
        adapter (ProductionSystemAdapter): Production system configuration

    Raises:
        ValueError: If multiple objects are positioned at the same location.
    """
    machine_locations = []
    for production_resource in get_production_resources(adapter):
        machine_locations.append(production_resource.location)
    source_locations = []
    for source in adapter.source_data:
        source_locations += get_location_of_locatable(adapter, source)
    source_locations = remove_duplicate_locations(source_locations)
    sink_locations = []
    for sink in adapter.sink_data:
        sink_locations += get_location_of_locatable(adapter, sink)
    sink_locations = remove_duplicate_locations(
        [sink.location for sink in adapter.sink_data]
    )
    store_locations = []
    for store in adapter.port_data:
        if not isinstance(store, port_data.StoreData):
            continue
        store_locations += get_location_of_locatable(adapter, store)

    positions = machine_locations + store_locations # + source_locations + sink_locations
    for location in positions:
        if positions.count(location) > 1:
            raise ValueError(
                f"Multiple objects are positioned at the same location: {location}"
            )


def assert_all_links_available(adapter: ProductionSystemData):
    """
    Asserts that all links are valid, so that the start and target of the link are valid locations.

    Args:
        adapter (ProductionSystemAdapter): Production system configuration

    Raises:
        ValueError: If the start or target of a link is not a valid location.
    """
    link_transport_processes = [
        process
        for process in adapter.process_data
        if isinstance(process, processes_data_module.LinkTransportProcessData)
    ]
    if not link_transport_processes:
        return
    nodes = get_set_of_IDs(adapter.node_data)
    resources = get_set_of_IDs(adapter.resource_data)
    sources = get_set_of_IDs(adapter.source_data)
    sinks = get_set_of_IDs(adapter.sink_data)
    ports = get_set_of_IDs(adapter.port_data)
    all_location_ids = nodes | resources | sources | sinks | ports
    for link_transport_process in link_transport_processes:
        link_pairs: List[List[str]] = []
        if isinstance(link_transport_process.links, dict):
            for start, targets in link_transport_process.links.items():
                for target in targets:
                    link_pairs.append([start, target])
        else:
            link_pairs = link_transport_process.links
        for start, target in link_pairs:
            if start not in all_location_ids:
                raise ValueError(
                    f"The link from {start} to {target} of process {link_transport_process.ID} is not a valid location because {start} is no valid location id."
                )
            if target not in all_location_ids:
                raise ValueError(
                    f"The link from {start} to {target} of process {link_transport_process.ID} is not a valid location because {target} is no valid location id."
                )


def assert_ports_have_locations(adapter: ProductionSystemData):
    """
    Checks if ports associated with resources have locations.
    Ports without locations cannot be used for transport routing.
    
    Note: Source and sink ports (SourceQueue, SinkQueue) are excluded from this check
    as they are logical grouping points that may be shared across multiple sources/sinks.

    Args:
        adapter (ProductionSystemAdapter): Adapter containing the production system to validate.

    Raises:
        ValueError: If resource ports are missing locations.
    """
    # Only check ports that are referenced by RESOURCES (not sources/sinks)
    # Sources and sinks use shared SourceQueue/SinkQueue which don't need individual locations
    resource_port_ids = set()
    for resource in adapter.resource_data:
        if resource.ports:
            resource_port_ids.update(resource.ports)
    
    # Check that all resource ports have locations
    ports_without_locations = []
    for port in adapter.port_data:
        if port.ID in resource_port_ids and port.location is None:
            # Skip common shared queues
            if port.ID in ['SourceQueue', 'SinkQueue']:
                continue
            ports_without_locations.append(port.ID)
    
    if ports_without_locations:
        raise ValueError(
            f"The following resource ports are missing locations (required for transport routing): {ports_without_locations}. "
            f"Ports must have locations to enable transport compatibility precomputation. "
            f"Use add_default_queues_to_resources() to automatically set port locations from resource locations."
        )


def check_for_clean_compound_processes(
    adapter_object: ProductionSystemData,
) -> bool:
    """
    Checks that the compound processes are clean, i.e. that they do not contain compund processes and normal processes at the same time.

    Args:
        adapter_object (ProductionSystemAdapter): Production system configuration

    Returns:
        bool: True if the compound processes are clean, False otherwise
    """
    possible_production_processes_ids = get_possible_production_processes_IDs(
        adapter_object
    )
    if any(
        isinstance(process_id, tuple)
        for process_id in possible_production_processes_ids
    ) and any(
        isinstance(process_id, str) for process_id in possible_production_processes_ids
    ):
        return False
    return True


def get_possible_production_processes_IDs(
    adapter_object: ProductionSystemData,
) -> Union[List[str], List[Tuple[str, ...]]]:
    """
    Returns all possible production processes IDs that can be used in the production system.
    Compund processes are grouped as touples, whereas individual processes are represented as strings.

    Args:
        adapter_object (ProductionSystemAdapter): Production system configuration

    Returns:
        Union[List[str], List[Tuple[str, ...]]]: List of production process IDs
    """
    possible_processes = [
        process
        for process in adapter_object.process_data
        if not isinstance(process, processes_data_module.TransportProcessData)
        and not isinstance(process, processes_data_module.RequiredCapabilityProcessData)
    ]
    compund_processes = [
        process
        for process in adapter_object.process_data
        if isinstance(process, processes_data_module.CompoundProcessData)
    ]
    compound_process_id_tuples = [
        tuple(compound_process.process_ids) for compound_process in compund_processes
    ]

    compound_processes_ids = set(
        [compound_process.ID for compound_process in compund_processes]
    )
    compound_processes_contained_process_ids = set(
        util.flatten(compound_process_id_tuples)
    )
    individual_processes_ids = [
        process.ID
        for process in possible_processes
        if process.ID not in compound_processes_contained_process_ids
        and process.ID not in compound_processes_ids
    ]

    return individual_processes_ids + compound_process_id_tuples


def get_possible_transport_processes_IDs(
    adapter_object: ProductionSystemData,
) -> List[str]:
    possible_processes = adapter_object.process_data
    return [
        process.ID
        for process in possible_processes
        if isinstance(process, processes_data_module.TransportProcessData)
    ]


def get_production_processes_from_ids(
    adapter_object: ProductionSystemData, process_ids: List[str]
) -> List[processes_data_module.PROCESS_DATA_UNION]:
    processes = []
    for process_id in process_ids:
        for process in adapter_object.process_data:
            if process.ID == process_id and (
                isinstance(process, processes_data_module.ProductionProcessData)
                or isinstance(process, processes_data_module.ReworkProcessData)
            ):
                processes.append(process)
                break
    return processes


def get_transport_processes_from_ids(
    adapter_object: ProductionSystemData, process_ids: List[str]
) -> List[processes_data_module.PROCESS_DATA_UNION]:
    processes = []
    for process_id in process_ids:
        for process in adapter_object.process_data:
            if (
                process.ID == process_id
                and isinstance(process, processes_data_module.TransportProcessData)
                and not (
                    hasattr(process, "capability") and getattr(process, "capability")
                )
            ):
                processes.append(process)
                break
    return processes


def get_capability_processes_from_ids(
    adapter_object: ProductionSystemData, process_ids: List[str]
) -> List[processes_data_module.PROCESS_DATA_UNION]:
    processes = []
    for process_id in process_ids:
        for process in adapter_object.process_data:
            if process.ID == process_id and (
                isinstance(process, processes_data_module.CapabilityProcessData)
                or (hasattr(process, "capability") and getattr(process, "capability"))
            ):
                processes.append(process)
                break
    return processes


def get_compound_processes_from_ids(
    adapter_object: ProductionSystemData, process_ids: List[str]
) -> List[processes_data_module.PROCESS_DATA_UNION]:
    processes = []
    for process_id in process_ids:
        for process in adapter_object.process_data:
            if process.ID == process_id and isinstance(
                process, processes_data_module.CompoundProcessData
            ):
                processes.append(process)
    return processes


def get_required_capability_processes_from_ids(
    adapter_object: ProductionSystemData, process_ids: List[str]
) -> List[processes_data_module.PROCESS_DATA_UNION]:
    processes = []
    for process_id in process_ids:
        for process in adapter_object.process_data:
            if process.ID == process_id and isinstance(
                process, processes_data_module.RequiredCapabilityProcessData
            ):
                processes.append(process)
    return processes


def get_contained_production_processes_from_compound_processes(
    adapter_object: ProductionSystemData,
    compound_processes: List[processes_data_module.CompoundProcessData],
) -> List[processes_data_module.PROCESS_DATA_UNION]:
    processes = []
    for compound_process in compound_processes:
        for process_id in compound_process.process_ids:
            processes = get_production_processes_from_ids(adapter_object, [process_id])
            if len(processes) > 1:
                raise ValueError(f"Multiple processes with ID {process_id} found.")
            if processes:
                process = processes[0]
                processes.append(process)
    return processes


def get_contained_capability_processes_from_compound_processes(
    adapter_object: ProductionSystemData,
    compound_processes: List[processes_data_module.CompoundProcessData],
) -> List[processes_data_module.PROCESS_DATA_UNION]:
    processes = []
    for compound_process in compound_processes:
        for process_id in compound_process.process_ids:
            processes = get_capability_processes_from_ids(adapter_object, [process_id])
            if len(processes) > 1:
                raise ValueError(f"Multiple processes with ID {process_id} found.")
            if processes:
                process = processes[0]
                processes.append(process)
    return processes


def get_contained_transport_processes_from_compound_processes(
    adapter_object: ProductionSystemData,
    compound_processes: List[processes_data_module.CompoundProcessData],
) -> List[processes_data_module.PROCESS_DATA_UNION]:
    processes = []
    for compound_process in compound_processes:
        for process_id in compound_process.process_ids:
            processes = get_transport_processes_from_ids(adapter_object, [process_id])
            if len(processes) > 1:
                raise ValueError(f"Multiple processes with ID {process_id} found.")
            if processes:
                process = processes[0]
                processes.append(process)
    return processes


def get_contained_required_capability_processes_from_compound_processes(
    adapter_object: ProductionSystemData,
    compound_processes: List[processes_data_module.CompoundProcessData],
) -> List[processes_data_module.PROCESS_DATA_UNION]:
    processes = []
    for compound_process in compound_processes:
        for process_id in compound_process.process_ids:
            processes = get_required_capability_processes_from_ids(
                adapter_object, [process_id]
            )
            if len(processes) > 1:
                raise ValueError(f"Multiple processes with ID {process_id} found.")
            process = processes[0]
            processes.append(process)
    return processes


def get_required_process_ids(
    configuration: ProductionSystemData,
) -> List[str]:
    """
    Returns all required process IDs that are used in the production system.

    Args:
        configuration (ProductionSystemData): Production system configuration

    Returns:
        List[str]: List of required process IDs
    """
    required: List[str] = []
    required_dict_keys: List[str] = []
    required_dict_values: List[str] = []

    for product in configuration.product_data:
        processes = product.processes
        if processes is None:
            continue
        if isinstance(processes, dict):
            required_dict_keys.extend(processes.keys())
            dict_values = util.flatten(list(processes.values()))
            if dict_values:
                required_dict_values.extend(dict_values)
        else:
            flattened = util.flatten(processes)
            if flattened:
                required.extend(flattened)

    required_transport_processes = [
        product.transport_process
        for product in configuration.product_data
        if product.transport_process
    ]

    primitive_transport_processes = [
        primitive.transport_process
        for primitive in configuration.primitive_data
        if getattr(primitive, "transport_process", None)
    ]

    dependency_entries = getattr(configuration, "dependency_data", None)
    if dependency_entries is None:
        dependency_entries = []
    dependency_processes = [
        dependency.required_process
        for dependency in dependency_entries
        if hasattr(dependency, "required_process") and dependency.required_process
    ]

    return list(
        set(
            required
            + required_dict_keys
            + required_dict_values
            + required_transport_processes
            + primitive_transport_processes
            + dependency_processes
        )
    )


def get_available_process_ids(
    configuration: ProductionSystemData,
) -> List[str]:
    """
    Returns all available process IDs that are used in the production system.

    Args:
        configuration (ProductionSystemData): Production system configuration

    Returns:
        List[str]: List of available process IDs
    """
    return list(
        set(
            util.flatten(
                [resource.process_ids for resource in configuration.resource_data]
            )
        )
    )


def get_required_production_processes(
    configuration: ProductionSystemData,
) -> List[processes_data_module.PROCESS_DATA_UNION]:
    """
    Returns all required production processes in the production system.

    Args:
        configuration (ProductionSystemData): Production system configuration

    Returns:
        List: List of required production processes
    """
    required = get_required_process_ids(configuration)
    required_production_processes = get_production_processes_from_ids(
        configuration, required
    )
    all_process_ids = set([process.ID for process in configuration.process_data])
    compound_processes = get_compound_processes_from_ids(configuration, all_process_ids)
    for compound_process in compound_processes:
        if not all(
            process_id in all_process_ids for process_id in compound_process.process_ids
        ):
            raise ValueError(
                f"Compound process {compound_process.ID} contains processes that are not available in the data."
            )
    required_compound_processes = get_compound_processes_from_ids(
        configuration, required
    )
    required_production_processes += (
        get_contained_production_processes_from_compound_processes(
            configuration, required_compound_processes
        )
    )
    return required_production_processes


def get_required_transport_processes(
    configuration: ProductionSystemData,
) -> List[processes_data_module.PROCESS_DATA_UNION]:
    """
    Returns all required transport processes in the production system.

    Args:
        configuration (ProductionSystemData): Production system configuration

    Returns:
        List: List of required transport processes
    """
    required = get_required_process_ids(configuration)
    required_transport_processes = get_transport_processes_from_ids(
        configuration, required
    )
    all_process_ids = set([process.ID for process in configuration.process_data])
    compound_processes = get_compound_processes_from_ids(configuration, all_process_ids)
    required_compound_processes = get_compound_processes_from_ids(
        configuration, required
    )
    required_transport_processes += (
        get_contained_transport_processes_from_compound_processes(
            configuration, required_compound_processes
        )
    )
    return required_transport_processes

def get_conveyor_processes(
    configuration: ProductionSystemData,
) -> List[processes_data_module.PROCESS_DATA_UNION]:
    """
    Returns all required transport processes in the production system.

    Args:
        configuration (ProductionSystemData): Production system configuration

    Returns:
        List: List of required transport processes
    """
    required_transport_processes=[]
    for process in configuration.process_data:
        if isinstance(process, LinkTransportProcessData):
            if not process.can_move:
                required_transport_processes.append(process)

    return required_transport_processes 

def get_required_capability_processes(
    configuration: ProductionSystemData,
) -> List[processes_data_module.PROCESS_DATA_UNION]:
    """
    Returns all required capability processes in the production system.

    Args:
        configuration (ProductionSystemData): Production system configuration

    Returns:
        List: List of required capability processes
    """
    required = get_required_process_ids(configuration)
    required_capability_processes = get_capability_processes_from_ids(
        configuration, required
    )
    required_capability_processes += get_required_capability_processes_from_ids(
        configuration, required
    )
    all_process_ids = set([process.ID for process in configuration.process_data])
    compound_processes = get_compound_processes_from_ids(configuration, all_process_ids)
    required_compound_processes = get_compound_processes_from_ids(
        configuration, required
    )
    required_capability_processes += (
        get_contained_capability_processes_from_compound_processes(
            configuration, required_compound_processes
        )
    )
    required_capability_processes += (
        get_contained_required_capability_processes_from_compound_processes(
            configuration, required_compound_processes
        )
    )
    return required_capability_processes


def get_available_production_processes(
    configuration: ProductionSystemData,
) -> List[processes_data_module.PROCESS_DATA_UNION]:
    """
    Returns all available production processes in the production system.

    Args:
        configuration (ProductionSystemData): Production system configuration

    Returns:
        List: List of available production processes
    """
    available = get_available_process_ids(configuration)
    available_production_processes = get_production_processes_from_ids(
        configuration, available
    )
    all_process_ids = set([process.ID for process in configuration.process_data])
    compound_processes = get_compound_processes_from_ids(configuration, all_process_ids)
    available_compound_processes = get_compound_processes_from_ids(
        configuration, available
    )
    available_production_processes += (
        get_contained_production_processes_from_compound_processes(
            configuration, available_compound_processes
        )
    )
    return available_production_processes


def get_available_transport_processes(
    configuration: ProductionSystemData,
) -> List[processes_data_module.PROCESS_DATA_UNION]:
    """
    Returns all available transport processes in the production system.

    Args:
        configuration (ProductionSystemData): Production system configuration

    Returns:
        List: List of available transport processes
    """
    available = get_available_process_ids(configuration)
    available_transport_processes = get_transport_processes_from_ids(
        configuration, available
    )
    all_process_ids = set([process.ID for process in configuration.process_data])
    compound_processes = get_compound_processes_from_ids(configuration, all_process_ids)
    available_compound_processes = get_compound_processes_from_ids(
        configuration, available
    )
    available_transport_processes += (
        get_contained_transport_processes_from_compound_processes(
            configuration, available_compound_processes
        )
    )
    return available_transport_processes


def get_available_capability_processes(
    configuration: ProductionSystemData,
) -> List[processes_data_module.PROCESS_DATA_UNION]:
    """
    Returns all available capability processes in the production system.

    Args:
        configuration (ProductionSystemData): Production system configuration

    Returns:
        List: List of available capability processes
    """
    available = get_available_process_ids(configuration)
    available_capability_processes = get_capability_processes_from_ids(
        configuration, available
    )
    all_process_ids = set([process.ID for process in configuration.process_data])
    compound_processes = get_compound_processes_from_ids(configuration, all_process_ids)
    available_compound_processes = get_compound_processes_from_ids(
        configuration, available
    )
    available_capability_processes += (
        get_contained_capability_processes_from_compound_processes(
            configuration, available_compound_processes
        )
    )
    return available_capability_processes


def get_missing_production_processes(
    available: List[
        Union[
            processes_data_module.ProductionProcessData,
            processes_data_module.ReworkProcessData,
        ]
    ],
    required: List[
        Union[
            processes_data_module.ProductionProcessData,
            processes_data_module.ReworkProcessData,
        ]
    ],
) -> List[Union[processes_data_module.ProductionProcessData, processes_data_module.ReworkProcessData]]:
    """
    Returns list of missing production processes.

    Args:
        available (List[processes_data_module.ProductionProcessData]): production processes that are available in the production system resources
        required (List[processes_data_module.ProductionProcessData]): production processes that are required from the products
    
    Returns:
        List: List of missing production processes
    """
    available_ids = set([process.ID for process in available])
    required_ids = set([process.ID for process in required])
    missing = required_ids - available_ids
    missing_processes = [p for p in required if p.ID in missing]
    return missing_processes


def assert_production_processes_available(
    available: List[
        Union[
            processes_data_module.ProductionProcessData,
            processes_data_module.ReworkProcessData,
        ]
    ],
    required: List[
        Union[
            processes_data_module.ProductionProcessData,
            processes_data_module.ReworkProcessData,
        ]
    ],
):
    """
    Checks if all required production processes are available.

    Args:
        available (List[processes_data_module.ProductionProcessData]): production processes that are available in the production system resources
        required (List[processes_data_module.ProductionProcessData]): production processes that are required from the products
    Raises:
        ValueError: If required production processes are not available
    """
    available = set([process.ID for process in available])
    required = set([process.ID for process in required])
    if required - available != set():
        raise ValueError(
            f"Required production processes {required - available} are not available."
        )


def get_missing_transport_processes(
    available: List[processes_data_module.TransportProcessData],
    required: List[processes_data_module.TransportProcessData],
) -> List[processes_data_module.TransportProcessData]:
    """
    Returns list of missing transport processes.

    Args:
        available (List[processes_data_module.TransportProcessData]): transport processes that are available in the production system resources
        required (List[processes_data_module.TransportProcessData]): transport processes that are required from the products

    Returns:
        List: List of missing transport processes
    """
    available_ids = set([process.ID for process in available])
    required_ids = set([process.ID for process in required])
    missing = required_ids - available_ids
    missing_processes = [p for p in required if p.ID in missing]
    return missing_processes


def assert_transport_processes_available(
    available: List[processes_data_module.TransportProcessData],
    required: List[processes_data_module.TransportProcessData],
):
    """
    Checks if all required transport processes are available.

    Args:
        available (List[processes_data_module.TransportProcessData]): transport processes that are available in the production system resources
        required (List[processes_data_module.TransportProcessData]): transport processes that are required from the products

    Raises:
        ValueError: If required transport processes are not available
    """
    available = set([process.ID for process in available])
    required = set([process.ID for process in required])
    if required - available != set():
        raise ValueError(
            f"Required transport processes {required - available} are not available."
        )


def get_missing_capability_processes(
    available: List[processes_data_module.CapabilityProcessData],
    required: List[processes_data_module.CapabilityProcessData],
) -> List[processes_data_module.CapabilityProcessData]:
    """
    Returns list of missing capability processes.

    Args:
        available (List[processes_data_module.CapabilityProcessData]): capability processes that are available in the production system resources
        required (List[processes_data_module.CapabilityProcessData]): capability processes that are required from the products

    Returns:
        List: List of missing capability processes
    """
    available_ids = set([process.capability for process in available])
    required_ids = set([process.capability for process in required])
    missing_capabilities = required_ids - available_ids
    missing_processes = [
        p
        for p in required
        if p.capability in missing_capabilities
    ]
    return missing_processes


def assert_capability_processes_available(
    available: List[processes_data_module.CapabilityProcessData],
    required: List[processes_data_module.CapabilityProcessData],
):
    """
    Checks if all required capability processes are available.

    Args:
        available (List[processes_data_module.CapabilityProcessData]): capability processes that are available in the production system resources
        required (List[processes_data_module.CapabilityProcessData]): capability processes that are required from the products

    Raises:
        ValueError: If required capability processes are not available
    """
    available = set([process.capability for process in available])
    required = set([process.capability for process in required])
    if required - available != set():
        raise ValueError(
            f"Required capability processes {required - available} are not available."
        )


def assert_required_processes_in_resources_available(
    configuration: ProductionSystemData,
):
    """
    Asserts that all required processes are available in the resources that are requested by the products in the configuration.

    Args:
        configuration (ProductionSystemAdapter): Production system configuration

    Raises:
        ValueError: If specified processes contain some logical errors.
    """
    process_map = {process.ID: process for process in configuration.process_data}

    available = set(
        util.flatten(
            [resource.process_ids for resource in configuration.resource_data]
        )
    )

    required_ids = set(get_required_process_ids(configuration))

    state_related_process_ids: Set[str] = set()
    for state in configuration.state_data:
        if hasattr(state, "origin_setup") and state.origin_setup:
            state_related_process_ids.add(state.origin_setup)
        if hasattr(state, "target_setup") and state.target_setup:
            state_related_process_ids.add(state.target_setup)
        if hasattr(state, "process_id") and state.process_id:
            state_related_process_ids.add(state.process_id)
    required_ids.update(state_related_process_ids)

    required_list = list(required_ids)
    available_list = list(available)

    required_production_processes = get_production_processes_from_ids(
        configuration, required_list
    )
    required_transport_processes = get_transport_processes_from_ids(
        configuration, required_list
    )
    required_capability_processes = get_capability_processes_from_ids(
        configuration, required_list
    )
    required_capability_processes += get_required_capability_processes_from_ids(
        configuration, required_list
    )
    available_production_processes = get_production_processes_from_ids(
        configuration, available_list
    )
    available_transport_processes = get_transport_processes_from_ids(
        configuration, available_list
    )
    available_capability_processes = get_capability_processes_from_ids(
        configuration, available_list
    )
    available_required_capability_processes = (
        get_required_capability_processes_from_ids(configuration, available_list)
    )
    if available_required_capability_processes:
        raise ValueError(
            f"Required capability processes {available_required_capability_processes} should not be used for resources since no time model is given."
        )

    all_process_ids = set([process.ID for process in configuration.process_data])
    compound_processes = get_compound_processes_from_ids(configuration, all_process_ids)
    for compound_process in compound_processes:
        if not all(
            process_id in all_process_ids for process_id in compound_process.process_ids
        ):
            raise ValueError(
                f"Compound process {compound_process.ID} contains processes that are not available in the data."
            )
    required_compound_processes = get_compound_processes_from_ids(
        configuration, required_list
    )
    available_compound_processes = get_compound_processes_from_ids(
        configuration, available_list
    )

    required_production_processes += (
        get_contained_production_processes_from_compound_processes(
            configuration, required_compound_processes
        )
    )
    required_transport_processes += (
        get_contained_transport_processes_from_compound_processes(
            configuration, required_compound_processes
        )
    )
    required_capability_processes += (
        get_contained_capability_processes_from_compound_processes(
            configuration, required_compound_processes
        )
    )
    required_capability_processes += (
        get_contained_required_capability_processes_from_compound_processes(
            configuration, required_compound_processes
        )
    )
    available_production_processes += (
        get_contained_production_processes_from_compound_processes(
            configuration, available_compound_processes
        )
    )
    available_transport_processes += (
        get_contained_transport_processes_from_compound_processes(
            configuration, available_compound_processes
        )
    )
    available_capability_processes += (
        get_contained_capability_processes_from_compound_processes(
            configuration, available_compound_processes
        )
    )

    required_capabilities: Set[str] = {
        process.capability
        for process in required_capability_processes
        if getattr(process, "capability", None)
    }

    for resource in configuration.resource_data:
        for process_id in resource.process_ids:
            process_data = process_map.get(process_id)
            if process_data is None:
                raise ValueError(
                    f"The process {process_id} assigned to resource {resource.ID} is not defined in process data."
                )
            capability = getattr(process_data, "capability", None)
            if isinstance(
                process_data, processes_data_module.CapabilityProcessData
            ):
                if not capability or capability not in required_capabilities:
                    raise ValueError(
                        f"The capability process {process_id} of resource {resource.ID} provides capability "
                        f"{capability}, which is not required by any product."
                    )
                continue
            if isinstance(
                process_data, processes_data_module.RequiredCapabilityProcessData
            ):
                raise ValueError(
                    f"Resource {resource.ID} contains required capability process {process_id}. "
                    "Required capability processes are only allowed on products."
                )
            if capability:
                if capability not in required_capabilities:
                    raise ValueError(
                        f"The process {process_id} of resource {resource.ID} provides capability "
                        f"{capability}, which is not required by any product."
                    )
                continue
            if process_id not in required_ids:
                continue

    assert_production_processes_available(
        available_production_processes, required_production_processes
    )
    assert_transport_processes_available(
        available_transport_processes, required_transport_processes
    )
    assert_capability_processes_available(
        available_capability_processes, required_capability_processes
    )



import prodsys
from prodsys.models import production_system_data, resource_data, sink_data, source_data, node_data, port_data
from prodsys.models.production_system_data import get_production_resources
from prodsys.util.node_link_generation.table_configuration import TableConfiguration
from prodsys.util.node_link_generation.table_configuration import StationConfiguration
from prodsys.util.node_link_generation.table_configuration import Visualization
from prodsys.util.node_link_generation.table_configuration_nodes_edges import NodeEdgeGenerator
from prodsys.util.node_link_generation.edge_directionality import EdgeDirectionality 
from prodsys.util.node_link_generation.configuration import Configuration 
#import a_star_algorithm 
import networkx as nx
import prodsys.util.node_link_generation.format_to_networkx as format_to_networkx
from typing import List, Any, Set, Optional, Tuple, Union
import xml.etree.ElementTree as ET

def parse_drawio_rectangles(xml_path):
    tree = ET.parse(xml_path)
    root = tree.getroot()

    tables = []

    # Find all mxCell elements with vertex="1"
    for cell in root.iter("mxCell"):
        if cell.get("vertex") != "1":
            continue

        # Identify rotation -> reject non-zero (draw.io writes e.g. rotation=45)
        style = cell.get("style", "")
        # draw.io stores rotation as "rotation=xx"
        rotation = None
        for part in style.split(";"):
            if part.startswith("rotation="):
                rotation = float(part.split("=")[1])
        if rotation not in (None, 0):
            raise ValueError(f"Non-right-angle rotation detected in cell {cell.get('id')}.")

        geom = cell.find("mxGeometry")
        if geom is None:
            continue  # useless entry

        # draw.io omits coordinates if they're zero; default to 0
        x = float(geom.get("x", 0))
        y = float(geom.get("y", 0))
        w = float(geom.get("width", 0))
        h = float(geom.get("height", 0))

        # compute rectangle corner points
        tableXMin = x
        tableYMin = y
        tableXMax = x + w
        tableYMax = y + h

        rectangle = {
            "corner_nodes": [
                {"pose": [tableXMin, tableYMin, 0]},
                {"pose": [tableXMax, tableYMin, 0]},
                {"pose": [tableXMax, tableYMax, 0]},
                {"pose": [tableXMin, tableYMax, 0]},
            ],
            "center_node": [
                {"pose": [
                    (tableXMin + tableXMax) / 2,
                    (tableYMin + tableYMax) / 2,
                    0,
                ]}
            ]
        }

        tables.append(rectangle)

    return tables

def get_all_locations(productionsystem: production_system_data):
    locations = []
    for node in list(productionsystem.resource_data) + list(productionsystem.source_data) + list(productionsystem.sink_data): #list(productionsystem.port_data) + list(productionsystem.resource_data): #get all port locations
        locations.append([node.ID, [x for x in node.location]])
    return locations

def generate_stations_config(stations: list) -> None:
    """
    Function is used to generate a JSON file for the station configuration.
    """
    # Prepare the new data structure.
    new_data = {
        "physical_objects": [],
    }
    Trajectory_Node_Distance = 0
    # Convert the original data into the new format.
    for station in stations:
        # Unidirectional station. Has one trajectory node. In and out port are at the same side.
        # Pose is the center of the station defined by the bounding box.
        if station[4] == "U":
            station_data = {
                "base": "if2",
                "pose": [station[0], station[1], station[2]],
                "bounding_box": [[-station[3]/2, -station[3]/2, 0], [station[3]/2, station[3]/2, 0]],
                "poi": {"in_port.trajectory_node": [station[3]/2 + Trajectory_Node_Distance, 0, 0],
                        "out_port.trajectory_node": [station[3]/2 + Trajectory_Node_Distance, 0, 0]},
                "urdf_file": {
                    "_comment": "just an example, not correct file!!",
                    "urdf_file_url": "https://backend.fac6310.btia.emea.ide.rb/cadfiles/ClipsCover/MO_0037_Press_EPSdp/MO_0037_Press_EPSdp.obj"
                }
            }
        elif station[4] == "B":
            # Bidirectional station. Has two trajectory nodes. In and out port are at the opposite sides.
            # Pose is in the center of the station defined by the bounding box.
            station_data = {
                "base": "if2",
                "pose": [station[0], station[1], station[2]],
                "bounding_box": [[-station[3]/2, -station[3]/2, 0],
                                    [station[3]/2, station[3]/2, 0]],
                "poi": {"in_port.trajectory_node": [station[3]/2 + Trajectory_Node_Distance, 0, 0],
                        "out_port.trajectory_node": [-station[3]/2 - Trajectory_Node_Distance, 0, 0]},
                "urdf_file": {
                    "_comment": "just an example, not correct file!!",
                    "urdf_file_url": "https://backend.fac6310.btia.emea.ide.rb/cadfiles/ClipsCover/MO_0037_Press_EPSdp/MO_0037_Press_EPSdp.obj"
                }
            }
        new_data["physical_objects"].append(station_data)

    return new_data

def find_borders(productionsystem: production_system_data):
    #finds the area within all stations are located

    stations = get_all_locations(productionsystem)
    max_x, max_y, = stations[0][1]
    min_x, min_y = stations[0][1]
    for station in stations: #get the max x and y coordinates of the stations to determine the size of the table layout
        if station[1][0] > max_x:
            max_x = station[1][0]
        if station[1][1] > max_y:
            max_y = station[1][1]
        if station[1][0] < min_x:
            min_x = station[1][0]
        if station[1][1] < min_y:
            min_y = station[1][1]
    return min_x, min_y, max_x, max_y

def generator(productionsystem: production_system_data, area=None, visualize=False, style="grid", simple_connection=True) -> nx.Graph:
    # Generate tables and stations based on the production system layout.
    items = [
        [loc[1][0]] + [loc[1][1]] + [0] + [0] + ["U"]
        for loc in get_all_locations(productionsystem)
    ]
    stations = generate_stations_config(items)
    # Determine table configuration based on area or production system layout.
    if area is None:
        min_x, min_y, max_x, max_y = find_borders(productionsystem)
        # Calculate proportional margins based on production system size
        # Use 10% of range, but clamp between 1.0 and 10.0 units
        range_x = max_x - min_x
        range_y = max_y - min_y
        margin_x = max(range_x * 0.1, 1.0)
        margin_x = min(margin_x, 10.0)
        margin_y = max(range_y * 0.1, 1.0)
        margin_y = min(margin_y, 10.0)
        
        tableXMax = max_x + margin_x
        tableYMax = max_y + margin_y
        tableXMin = min_x - margin_x
        tableYMin = min_y - margin_y
        tables = {
            "corner_nodes": [
                {"pose": [tableXMin, tableYMin, 0]}, 
                {"pose": [tableXMax, tableYMin, 0]},
                {"pose": [tableXMax, tableYMax, 0]},
                {"pose": [tableXMin, tableYMax, 0]}
            ],
            "center_node": [
                {"pose": [((tableXMin + tableXMax)/2), ((tableYMin + tableYMax)/2), 0]}
            ]
        }
        # Get the dimensions of the TABLE layout (after adding margins), not the production system
        # This is important for calculating proper node spacing
        table_dim_x = tableXMax - tableXMin
        table_dim_y = tableYMax - tableYMin
        dim_x, dim_y = table_dim_x, table_dim_y
    else:
        tables = area
        min_x, min_y, _ = tables[0]['corner_nodes'][0]['pose']
        max_x, max_y, _ = tables[0]['corner_nodes'][0]['pose']
    
        for table in tables:
            for corner in table['corner_nodes']:
                if corner['pose'][0] > max_x:
                    max_x = corner['pose'][0]
                if corner['pose'][1] > max_y:
                    max_y = corner['pose'][1]
                if corner['pose'][0] < min_x:
                    min_x = corner['pose'][0]
                if corner['pose'][1] < min_y:
                    min_y = corner['pose'][1]
        # Get the dimensions of the table layout (correct calculation: range, not sum of absolutes)
        dim_x, dim_y = max_x - min_x, max_y - min_y
    
    # Set the dimensions of the layout.
    config = Configuration()
    config.set(Configuration.Dim_X, int(dim_x))
    config.set(Configuration.Dim_Y, int(dim_y))
    
    # CRITICAL FIX: Make hardcoded boundary/distance values proportional to table size
    # The default values (32, 24, 16) are designed for large tables (~100x50 units)
    # For small tables, these values prevent node generation. Scale them DOWN only for small tables.
    tablesize = min(dim_x, dim_y)
    
    # Store original values to restore later (Configuration is a singleton)
    original_min_node_dist = config.get(Configuration.Min_Node_Distance)
    original_min_edge_dist = config.get(Configuration.Min_Node_Edge_Distance)
    original_trajectory_dist = config.get(Configuration.Trajectory_Node_Distance)
    original_buffer_dist = config.get(Configuration.Buffer_Node_Distance)
    
    # Only scale DOWN for small tables (< 50 units). For large tables, use original values.
    # The original values work fine for large systems, but are too large for small systems.
    if tablesize < 50.0:
        # Scale down proportionally for small tables, but don't go below minimums
        scale_factor = max(tablesize / 50.0, 0.1)  # Scale based on 50 units as reference, minimum 0.1
        
        scaled_min_node_dist = max(int(original_min_node_dist * scale_factor), 1)  # At least 1 unit
        scaled_min_edge_dist = max(int(original_min_edge_dist * scale_factor), 1)
        scaled_trajectory_dist = max(int(original_trajectory_dist * scale_factor), 1)
        scaled_buffer_dist = max(int(original_buffer_dist * scale_factor), 1)
        
        config.set(Configuration.Min_Node_Distance, scaled_min_node_dist)
        config.set(Configuration.Min_Node_Edge_Distance, scaled_min_edge_dist)
        config.set(Configuration.Trajectory_Node_Distance, scaled_trajectory_dist)
        config.set(Configuration.Buffer_Node_Distance, scaled_buffer_dist)
    # For large tables (>= 50 units), keep original values - no scaling needed
    
    table_config = TableConfiguration(config)
    station_config = StationConfiguration(config, table_config)
    # Add tables and define corner nodes of the table configuration.
    node_edge_generator = NodeEdgeGenerator(config, table_config, station_config)
    edge_directionality = EdgeDirectionality(table_config, node_edge_generator)
    visualization = Visualization(config, table_config, station_config, node_edge_generator)
    networkx_formater = format_to_networkx.NetworkXGraphGenerator(node_edge_generator.graph, station_config)
    #vda5050_formater = format_to_vda5050.VDA5050JSONGenerator(node_edge_generator, station_config) #optionally allow 
    #a_star = a_star_algorithm.AStar(table_config, station_config, node_edge_generator, visualization)
    if isinstance(tables, list):
        for table in tables:  # in case of multiple tables
            table_config.add_table(table['corner_nodes'], visualization)
    else:
        table_config.add_table(tables['corner_nodes'], visualization)
    table_config.generate_table_configuration()

    # Add stations and update table configuration considering the stations.
    station_config.add_stations(stations, visualization)
    table_config.table_configuration_with_stations(station_config)
    #visualization.show_table_configuration(table_configuration=False, boundary=True, stations=True, station_nodes=True)

    # Define the medial axis of the table configuration.    MARKER: Optional
    if not simple_connection:    
        table_config.define_medial_axis()
    #visualization.show_table_configuration(table_configuration=True, boundary=True, stations=True, station_nodes=True, medial_axis=True)

    # Define zones of the table configuration.  MARKER: Optional
    if not simple_connection:
        table_config.define_zones()
    #visualization.show_table_configuration(table_configuration=True, boundary=True, stations=False, station_nodes=False, zones=True)

    # Define station nodes and edges. Add edges optionally.
    # Variable is defined, so it is consistent with functions called afterwards.
    # Station nodes (and edges) must be added before other nodes are added. Here only trajectory nodes are defined.
    add_edges = False
    node_edge_generator.add_station_nodes_and_edges(add_edges=add_edges, buffer_nodes=False) #this method is also used in add_outer_nodes_and_edges
    #visualization.show_table_configuration(table_configuration=False, boundary=False, stations=True, station_nodes=True, nodes=True, edges=True)

    # Add nodes (and edges) along the boundary of table configuration considering stations.
    # Intermediate nodes along the boundary edges can be added optionally. The minimum distance between nodes can be defined.
    add_nodes_between = True
    tablesize = min(dim_x, dim_y)

    if not simple_connection:
        min_node_distance = max(0.1*tablesize, 10)
        max_node_distance = 0.2*tablesize
    else:
        # For simple_connection, use proportional spacing but ensure enough nodes are generated
        # For very small tables (< 10 units), use a fixed small spacing to ensure multiple nodes
        # For larger tables, use proportional spacing
        if tablesize < 10.0:
            # Use spacing that ensures at least 4-5 nodes fit in each dimension
            min_node_distance = max(tablesize / 5.0, 0.3)  # At least 0.3, but aim for 5 nodes per dimension
            max_node_distance = max(tablesize / 4.0, 0.5)   # At least 0.5, but aim for 4 nodes per dimension
        else:
            min_node_distance = max(0.1*tablesize, 0.5)  # At least 0.5 units, typically 10% of table size
            max_node_distance = max(0.15*tablesize, 1.0)  # At least 1.0 units, typically 15% of table size
    # Check if we need to add outer nodes: either resources can move, or there are LinkTransportProcesses
    has_movable_resources = any(resource.can_move for resource in productionsystem.resource_data if isinstance(resource.control_policy, resource_data.TransportControlPolicy))
    # Check for LinkTransportProcessData - need to import it
    from prodsys.models import processes_data
    has_link_transport = any(isinstance(process, processes_data.LinkTransportProcessData) for process in productionsystem.process_data)
    
    if has_movable_resources or has_link_transport:
        node_edge_generator.add_outer_nodes_and_edges(edge_directionality, add_nodes_between=add_nodes_between, max_node_distance=max_node_distance, min_node_distance=min_node_distance, add_edges=add_edges)
    #visualization.show_table_configuration(table_configuration=False, boundary=False, stations=True, station_nodes=True, nodes=True, edges=True)

    # Define random nodes in the free space of the table configuration.
    # choose a grid generation method that is defined here
    if style == "grid":
        node_edge_generator.define_global_grid(grid_spacing=min_node_distance, adjust_spacing=False, add_corner_nodes_first=False)
    elif style == "random":
        node_edge_generator.define_random_nodes(min_node_distance=min_node_distance)
    #visualization.show_table_configuration(table_configuration=False, boundary=False, stations=True, station_nodes=True, nodes=True, edges=True)

    # Connect nodes by edges using the Delaunay triangulation.
    # Check if we have enough nodes (need at least 4 for Delaunay triangulation)
    num_nodes = len(node_edge_generator.graph.nodes)
    smaller_spacing = min_node_distance  # Initialize for error message
    if num_nodes < 4:
        # If we don't have enough nodes, try generating with progressively smaller spacing
        # This can happen with very small tables or when grid generation filters out too many nodes
        for attempt in range(3):  # Try up to 3 times with smaller spacing
            smaller_spacing = min_node_distance / (2 ** (attempt + 1))
            smaller_spacing = max(smaller_spacing, 0.1)  # Don't go below 0.1 units
            if style == "grid":
                node_edge_generator.define_global_grid(grid_spacing=smaller_spacing, adjust_spacing=False, add_corner_nodes_first=False)
            elif style == "random":
                node_edge_generator.define_random_nodes(min_node_distance=smaller_spacing)
            num_nodes = len(node_edge_generator.graph.nodes)
            if num_nodes >= 4:
                break
    
    if num_nodes < 4:
        # Last resort: try with very small fixed spacing
        if num_nodes < 4:
            very_small_spacing = 0.2  # Very small fixed spacing
            if style == "grid":
                node_edge_generator.define_global_grid(grid_spacing=very_small_spacing, adjust_spacing=True, add_corner_nodes_first=True)
            elif style == "random":
                node_edge_generator.define_random_nodes(min_node_distance=very_small_spacing)
            num_nodes = len(node_edge_generator.graph.nodes)
    
    if num_nodes < 4:
        raise ValueError(f"Not enough nodes generated ({num_nodes}) for Delaunay triangulation (need at least 4). "
                        f"Table size: {tablesize:.2f}, initial spacing: {min_node_distance:.2f}, final spacing tried: {smaller_spacing:.2f}. "
                        f"This may indicate an issue with the table configuration or node generation algorithm. "
                        f"Table bounds: X=[{tableXMin if 'tableXMin' in locals() else 'N/A':.2f}, {tableXMax if 'tableXMax' in locals() else 'N/A':.2f}], "
                        f"Y=[{tableYMin if 'tableYMin' in locals() else 'N/A':.2f}, {tableYMax if 'tableYMax' in locals() else 'N/A':.2f}]. "
                        f"Try using a larger table area, different style ('random' instead of 'grid'), or disabling simple_connection.")
    
    node_edge_generator.delaunay_triangulation(nodes=node_edge_generator.graph.nodes, without_distance_check=False)
    #visualization.show_table_configuration(table_configuration=False, boundary=False, stations=True, station_nodes=True, nodes=True, edges=True)

    # Define boundary nodes and edges. Nodes with 1 or 2 edges must be removed first.
    edge_directionality.define_boundary_nodes_and_edges()

    # Define boundary edges unidirectional (if possible). Graph connections must be updated afterwards.
    exterior_direction = 'ccw'
    edge_directionality.define_boundary_edges_directionality(exterior_direction=exterior_direction, narrow_sections_unidirectional=False)
    node_edge_generator.graph.update_graph_connections()
    if visualize:
        visualization.show_table_configuration(table_configuration=False, stations=True, station_nodes=False, nodes=True, edges=True)

    # Generate networkx graph. Bidirectional graph and mixed-directional graph are generated.
    # The graph can be visualized. 
    G, DiG = networkx_formater.generate_nx_graph(plot=False) #G: undirected graph, DiG: directed graph

    # Restore original Configuration values (Configuration is a singleton, so we need to restore)
    # to prevent test pollution when tests run in sequence
    if tablesize < 50.0:
        config.set(Configuration.Min_Node_Distance, original_min_node_dist)
        config.set(Configuration.Min_Node_Edge_Distance, original_min_edge_dist)
        config.set(Configuration.Trajectory_Node_Distance, original_trajectory_dist)
        config.set(Configuration.Buffer_Node_Distance, original_buffer_dist)

    return G

def convert_nx_to_prodsys(adapter: production_system_data, G: nx.Graph):

    #all_locations = get_all_locations(productionsystem)
    #all_relevant_resources = [prodres.ID for prodres in get_production_resources(productionsystem)]
    #all_relevant_resources.extend([sink.ID for sink in productionsystem.sink_data])
    #all_relevant_resources.extend([source.ID for source in productionsystem.source_data])

    # Include ALL resources (production + transport), not just production resources
    # This ensures transport resources like Workers_Part1 are also included in the network
    from prodsys.models.production_system_data import get_transport_resources
    all_locations = [(prodres.ID, prodres.location) for prodres in get_production_resources(adapter)]
    all_locations.extend([(tres.ID, tres.location) for tres in get_transport_resources(adapter)])
    all_locations.extend([(sink.ID, sink.location) for sink in adapter.sink_data])
    all_locations.extend([(source.ID, source.location) for source in adapter.source_data])

    # Build a lookup: location -> list of resources
    location_to_resources = {}
    for resource in all_locations:
        key = tuple(resource[1])
        location_to_resources.setdefault(key, []).append(resource[0])

    nx_to_location = {}
    # Map GML node id (int) to matched resource ID
    # Also track which nodes will be added to node_data (for validation)
    node_id_to_name = {}  # Maps NetworkX node ID -> node name that will be in node_data
    new_nodes = [] #nodes that where generated and not previously existing in the production system
    node_blocks = G.nodes
    for block in node_blocks: #extract the newly generated nodes from the networkx graph
        pos = G.nodes[block].get("pos")
        if len(pos) >= 2:
            x = pos[0] 
            y = pos[1] 
        else:
            raise ValueError(f"Node {block} position is not two-dimensional: {pos}") #Generator can only handle 2D positions

        nx_to_location[block] = pos
        node_name = f"node_{block}"
        if pos not in location_to_resources:
            # This node will be added to node_data
            new_nodes.append([node_name, (x, y)])
            node_id_to_name[block] = node_name
        else:
            # This node is at a resource location, so it won't be added as a separate node
            # But we still need to track it for linking purposes
            # In this case, links should use the resource ID, not the node ID
            node_id_to_name[block] = None  # Mark as not a separate node


    new_links = []
    edge_blocks = G.edges
    for block in edge_blocks:
        src_id, tgt_id = block[:2]
        src_loc, tgt_loc = nx_to_location[src_id], nx_to_location[tgt_id]
        if src_loc in location_to_resources and tgt_loc in location_to_resources:
            for src_resource in location_to_resources[src_loc]:
                for tgt_resource in location_to_resources[tgt_loc]:
                    new_links.append([src_resource, tgt_resource])
        elif src_loc in location_to_resources:
            # Source is a resource, target might be a node or resource
            if tgt_loc in location_to_resources:
                # Both are resources - should have been handled above
                pass
            else:
                # Target is a node - use node name if it will be in node_data
                tgt_name = node_id_to_name.get(tgt_id)
                if tgt_name:
                    for src_resource in location_to_resources[src_loc]:
                        new_links.append([src_resource, tgt_name])
                # If tgt_name is None, the node is at a resource location, so we should link to that resource
                if not tgt_name and tgt_loc in location_to_resources:
                    for src_resource in location_to_resources[src_loc]:
                        for tgt_resource in location_to_resources[tgt_loc]:
                            new_links.append([src_resource, tgt_resource])
        elif tgt_loc in location_to_resources:
            # Target is a resource, source might be a node or resource
            src_name = node_id_to_name.get(src_id)
            if src_name:
                for tgt_resource in location_to_resources[tgt_loc]:
                    new_links.append([src_name, tgt_resource])
            # If src_name is None, the node is at a resource location, so we should link to that resource
            if not src_name and src_loc in location_to_resources:
                for src_resource in location_to_resources[src_loc]:
                    for tgt_resource in location_to_resources[tgt_loc]:
                        new_links.append([src_resource, tgt_resource])
        else:
            # Both are nodes - create links appropriately
            src_name = node_id_to_name.get(src_id)
            tgt_name = node_id_to_name.get(tgt_id)
            
            if src_name and tgt_name:
                # Both nodes will be in node_data
                new_links.append([src_name, tgt_name])
            elif src_name and not tgt_name:
                # Source is a node, target is at a resource location
                if tgt_loc in location_to_resources:
                    for tgt_resource in location_to_resources[tgt_loc]:
                        new_links.append([src_name, tgt_resource])
            elif not src_name and tgt_name:
                # Source is at a resource location, target is a node
                if src_loc in location_to_resources:
                    for src_resource in location_to_resources[src_loc]:
                        new_links.append([src_resource, tgt_name])
            elif not src_name and not tgt_name:
                # Both nodes are at resource locations - link the resources
                if src_loc in location_to_resources and tgt_loc in location_to_resources:
                    for src_resource in location_to_resources[src_loc]:
                        for tgt_resource in location_to_resources[tgt_loc]:
                            new_links.append([src_resource, tgt_resource])

    # CRITICAL FIX: Ensure ALL resources/sources/sinks are connected
    # Some stations may not get nodes in the graph due to station configuration limitations
    # Connect any unconnected resources to the nearest node in the graph
    connected_resources = set()
    for link in new_links:
        connected_resources.add(link[0])
        connected_resources.add(link[1])
    
    # Find resources that aren't connected
    all_resource_ids = {r[0] for r in all_locations}
    unconnected_resources = {rid for rid in all_resource_ids if rid not in connected_resources}
    
    if unconnected_resources and nx_to_location:
        # Helper function to find nearest node
        def find_nearest_node_id(resource_loc):
            """Find the ID of the nearest node to the given location."""
            min_dist = float('inf')
            nearest_id = None
            for node_id, node_loc in nx_to_location.items():
                if len(node_loc) >= 2 and len(resource_loc) >= 2:
                    dx = node_loc[0] - resource_loc[0]
                    dy = node_loc[1] - resource_loc[1]
                    dist = (dx*dx + dy*dy) ** 0.5
                    if dist < min_dist:
                        min_dist = dist
                        nearest_id = node_id
            return nearest_id
        
        # Connect each unconnected resource to its nearest connected location
        # IMPORTANT: Only create links with IDs that will be valid (in node_data, resource_data, etc.)
        for resource_id in unconnected_resources:
            resource_loc = next((r[1] for r in all_locations if r[0] == resource_id), None)
            if resource_loc:
                # Find nearest connected resource first (preferred)
                nearest_connected_resource = None
                min_dist_to_resource = float('inf')
                for link in new_links:
                    for target in [link[0], link[1]]:
                        if target in all_resource_ids and target in connected_resources:
                            target_loc = next((r[1] for r in all_locations if r[0] == target), None)
                            if target_loc and len(target_loc) >= 2 and len(resource_loc) >= 2:
                                dx = target_loc[0] - resource_loc[0]
                                dy = target_loc[1] - resource_loc[1]
                                dist = (dx*dx + dy*dy) ** 0.5
                                if dist < min_dist_to_resource:
                                    min_dist_to_resource = dist
                                    nearest_connected_resource = target
                
                # If we found a connected resource, link to it
                if nearest_connected_resource:
                    new_links.append([resource_id, nearest_connected_resource])
                    new_links.append([nearest_connected_resource, resource_id])
                else:
                    # Fallback: find nearest node that will be in node_data
                    nearest_node_id = find_nearest_node_id(resource_loc)
                    if nearest_node_id is not None:
                        nearest_node_name = node_id_to_name.get(nearest_node_id)
                        if nearest_node_name:
                            # Node will be in node_data, so we can link to it
                            new_links.append([resource_id, nearest_node_name])
                            new_links.append([nearest_node_name, resource_id])
                        else:
                            # Node is at a resource location - link to that resource instead
                            nearest_node_loc = nx_to_location.get(nearest_node_id)
                            if nearest_node_loc and tuple(nearest_node_loc) in location_to_resources:
                                nearest_resource = location_to_resources[tuple(nearest_node_loc)][0]
                                if nearest_resource in connected_resources:
                                    new_links.append([resource_id, nearest_resource])
                                    new_links.append([nearest_resource, resource_id])

    return new_nodes, new_links

def apply_nodes_links(adapter: production_system_data, nodes, links) -> None:

    
    # Replace node_data in the Prodocutionsystem
    adapter.node_data = []
    for node in nodes:
        adapter.node_data.append(node_data.NodeData(ID=str(node[0]), description="", location=node[1]))

    # Replace links inside LinkTransportProcesses in the Productionsystem
    # Update links for ALL LinkTransportProcesses to ensure old links are cleared
    # For conveyors (can_move=False), we update all of them to handle multiple conveyor processes
    for LinkTransportProcess in adapter.process_data:
        if isinstance(LinkTransportProcess, prodsys.processes_data.LinkTransportProcessData):
            # Update links for all processes to clear any old/stale links
            LinkTransportProcess.links = links
            # Note: Original code broke after first conveyor, but we now update all conveyors
            # to handle cases with multiple conveyor processes in evolutionary algorithm

def generate_and_apply_network(adapter: production_system_data, xml_path = None, visualize=False, style="grid", simple_connection=True) -> None:
    if xml_path:
        tables = parse_drawio_rectangles(xml_path)
    else:
        tables = None
    G = generator(adapter, tables, visualize=visualize, style=style, simple_connection=simple_connection)
    nodes, links = convert_nx_to_prodsys(adapter, G)
    apply_nodes_links(adapter, nodes, links)

def get_new_links(adapter: production_system_data, style="grid", simple_connection=True) -> List[Tuple[str, str]]:
    G = generator(adapter, style=style, simple_connection=simple_connection)
    _, new_links = convert_nx_to_prodsys(adapter, G)
    return new_links


from __future__ import annotations

from typing import List, TYPE_CHECKING, Literal

from prodsys.simulation import port, sim
from prodsys.models import sink_data

if TYPE_CHECKING:
    from prodsys.simulation.entities import product


class Sink:
    """
    Class that represents a sink.

    Args:
        env (sim.Environment): The simulation environment.
        data (sink_data.SinkData): The sink data.
        product_factory (product_factory.ProductFactory): The product factory.
        input_queues (List[store.Queue], optional): The input queues. Defaults to [].
    """

    def __init__(
        self,
        env: sim.Environment,
        data: sink_data.SinkData,
        product_factory: product_factory.ProductFactory,
    ):
        self.env = env
        self.data = data
        self.product_factory = product_factory
        self.ports: List[port.Queue] = []
        self.can_move = False

    def add_ports(self, ports: List[port.Queue]):
        """
        Adds input queues to the sink.

        Args:
            ports (List[store.Queue]): The input ports.
        """
        self.ports.extend(ports)

    def get_location(self) -> List[float]:
        return self.data.location

    def get_input_queue_length(self) -> int:
        """
        Returns total number of items in all input_queues. Defaults to 0 for a sink.

        Returns:
            int(0)
        """
        return 0

    def get_output_queue_length(self) -> int:
        """
        Returns zero. Needed for transport control policies that use the target locations output queue for request ordering.
        Returns:
            int(0)
        """
        return 0

    def register_finished_product(self, product: product.Product):
        """
        Registers a finished product when it reaches the sink.

        Args:
            product (product.Product): The finished product.
        """
        router = self.product_factory.router

        self.product_factory.router.primitive_factory.primitives.append(product)
        if(product.data.becomes_consumable):    
            if self.product_factory.router:
                if product.data.type not in router.free_primitives_by_type:
                    router.free_primitives_by_type[product.data.type] = []
                router.free_primitives_by_type[product.data.type].append(product)
                # Notify all resources that might be waiting for this primitive type
                # This allows controllers to recheck feasibility of pending requests
                for resource in router.resources:
                    if resource.controller and not resource.controller.state_changed.triggered:
                        resource.controller.state_changed.succeed()
                if not router.got_primitive_request.triggered:
                    router.got_primitive_request.succeed()
                    
        self.product_factory.register_finished_product(product)


from prodsys.factories import product_factory



import prodsys.express as psx

import prodsys

# prodsys.set_logging("DEBUG")

t1 = psx.FunctionTimeModel("exponential", 0.8, 0, "t1")

p1 = psx.ProductionProcess(t1, "p1")

# t3 = psx.FunctionTimeModel("constant", 0.1, ID="t3")
t3 = psx.DistanceTimeModel(60, 0.05, "manhattan", ID="t3")

tp = psx.TransportProcess(t3, "tp")

machine = psx.Resource([p1], [5, 0], 1, ID="machine")

transport = psx.Resource([tp], [3, 0], 1, ID="transport")
transport2 = psx.Resource([tp], [4, 0], 1, ID="transport2")

storage1 = psx.Store(ID="storage1", location=[6, 0], capacity=30)
storage2 = psx.Store(ID="storage2", location=[11, 0], capacity=20)

workpiece_carrier_1 = psx.Primitive(
    ID="workpice_carrier_1",
    transport_process=tp,
    storages=[storage1, storage2],
    quantity_in_storages=[5, 20],
)

workpiece_carrier_dependency_1 = psx.ToolDependency(
    ID="workpiece_carrier_dependency_1",
    required_entity=workpiece_carrier_1,
)

product1 = psx.Product(
    process=[p1],
    transport_process=tp,
    ID="product1",
    dependencies=[workpiece_carrier_dependency_1],
)

sink1 = psx.Sink(product1, [10, 0], "sink1")

arrival_model_1 = psx.FunctionTimeModel("constant", 0.9, ID="arrival_model_1")

source1 = psx.Source(product1, arrival_model_1, [0, 0], ID="source_1")

system = psx.ProductionSystem([machine, transport, transport2], [source1], [sink1], [workpiece_carrier_1])

system.validate()
system.run(time_range=1000)
system.runner.save_results_as_csv()
system.runner.print_results()
# system.runner.plot_results()
# system.runner.plot_results_executive()


from __future__ import annotations

from typing import TYPE_CHECKING

# removed unused import P from pydantic.type_adapter

from prodsys.models import port_data
from prodsys.simulation import request

if TYPE_CHECKING:
    from prodsys.simulation import port


class InteractionHandler:

    def __init__(self):
        """
        Initializes the InteractionHandler.
        """
        # TODO: implement more sophisticated port selection heuristics (distance or number of items in the port based)
        self.port_selection_heuristic = random_port_selection_heuristic

        # TODO: precompute later the input and output ports of the resources, sinks and sources to be faster (use resource_factory, sink_factory, source_factory from router during init of InteractionHandler)
        self.input_ports_per_resource: dict[str, list[port.Queue]] = {}
        self.output_ports_per_resource: dict[str, list[port.Queue]] = {}

    def get_interaction_buffer(
        self, routed_request: request.Request
    ) -> port.Queue:
        if routed_request.request_type in (
            request.RequestType.ENTITY_DEPENDENCY,
            request.RequestType.PROCESS_DEPENDENCY,
            request.RequestType.RESOURCE_DEPENDENCY,
        ):
            return None
        elif routed_request.request_type == request.RequestType.TRANSPORT:
            return None
        elif routed_request.request_type in (request.RequestType.PRODUCTION, request.RequestType.PROCESS_MODEL):
            return self.handle_production_buffer(routed_request)
        else:
            raise ValueError(f"Unknown request type: {routed_request.request_type}")

    def handle_production_buffer(self, routed_request: request.Request) -> port.Queue:
        if not routed_request.resource.buffers:
            return None
        queue = self.port_selection_heuristic(routed_request.resource.buffers, routed_request)
        return queue

    def get_interaction_ports(
        self, routed_request: request.Request
    ) -> tuple[port.Queue, port.Queue]:
        """
        Returns the interaction ports for the given request.

        Args:
            request (request.Request): The request for which to get the interaction ports.

        Returns:
            tuple[store.Queue, store.Queue]: A tuple containing the input and output ports of the requesting item.
        """
        if routed_request.request_type in (
            request.RequestType.ENTITY_DEPENDENCY,
            request.RequestType.PROCESS_DEPENDENCY,
            request.RequestType.RESOURCE_DEPENDENCY,
        ):
            return None, None
        elif routed_request.request_type == request.RequestType.TRANSPORT:
            return self.handle_transport_interaction(routed_request)
        elif routed_request.request_type in (request.RequestType.PRODUCTION, request.RequestType.PROCESS_MODEL):
            return self.handle_production_interaction(routed_request)
        else:
            raise ValueError(f"Unknown request type: {routed_request.request_type}")

    def handle_transport_interaction(
        self, routed_request: request.Request
    ) -> tuple[port.Queue, port.Queue]:
        """
        Handles the interaction for transport requests.

        Args:
            routed_request (request.Request): The transport request.

        Returns:
            tuple[store.Queue, store.Queue]: A tuple containing the input and output ports of the requesting item.
        """
        # Get origin ports - handle Store objects that have store_ports instead of ports
        # For Node objects, we need to handle them specially since they don't have ports
        if hasattr(routed_request.origin, 'ports'):
            origin_ports = routed_request.origin.ports
        elif hasattr(routed_request.origin, 'store_ports'):
            origin_ports = routed_request.origin.store_ports
        else:
            # For Node objects or other locations without ports, return None to indicate direct location access
            origin_ports = None
            
        if origin_ports is None:
            # For Node objects or other locations without ports, use the location directly
            origin_port = routed_request.origin
        else:
            origin_output_ports = [
                q
                for q in origin_ports
                if q.data.interface_type
                in [
                    port_data.PortInterfaceType.OUTPUT,
                    port_data.PortInterfaceType.INPUT_OUTPUT,
                ]
            ]
            origin_port = self.get_origin_port(origin_output_ports, routed_request)

        # Get target ports - handle Store objects that have store_ports instead of ports
        if hasattr(routed_request.target, 'ports'):
            target_ports = routed_request.target.ports
        elif hasattr(routed_request.target, 'store_ports'):
            target_ports = routed_request.target.store_ports
        else:
            # For Node objects or other locations without ports, return None to indicate direct location access
            target_ports = None
            
        if target_ports is None:
            # For Node objects or other locations without ports, use the location directly
            target_port = routed_request.target
        else:
            target_input_ports = [
                q
                for q in target_ports
                if q.data.interface_type
                in [
                    port_data.PortInterfaceType.INPUT,
                    port_data.PortInterfaceType.INPUT_OUTPUT,
                ]
            ]
            # select target port that can accept the item
            target_port = self.get_target_port(target_input_ports, routed_request)

        return (origin_port, target_port)

    def handle_production_interaction(
        self, routed_request: request.Request
    ) -> tuple[port.Queue, port.Queue]:
        """
        Handles the interaction for production requests which input and output ports should be used at the resource.

        Args:
            routed_request (request.Request): The production request.

        Returns:
            tuple[store.Queue, store.Queue]: A tuple containing the input and output ports of the requesting item.
        """
        origin_input_ports = [
            q
            for q in routed_request.resource.ports
            if q.data.interface_type
            in [
                port_data.PortInterfaceType.INPUT,
                port_data.PortInterfaceType.INPUT_OUTPUT,
            ]
        ]
        origin_port = self.get_origin_port(origin_input_ports, routed_request)

        # Prefer dedicated OUTPUT ports for target; if none, use INPUT_OUTPUT but avoid the same
        # port selected for origin when possible to prevent returning to input.
        target_output_ports = [
            q
            for q in routed_request.resource.ports
            if q.data.interface_type
            in [
                port_data.PortInterfaceType.OUTPUT,
                port_data.PortInterfaceType.INPUT_OUTPUT,
            ]
        ]

        # Prioritize OUTPUT-only ports
        preferred_targets = [
            q for q in target_output_ports if q.data.interface_type == port_data.PortInterfaceType.OUTPUT
        ]
        fallback_targets = [
            q for q in target_output_ports if q.data.interface_type == port_data.PortInterfaceType.INPUT_OUTPUT and q is not origin_port
        ]
        filtered_targets = preferred_targets or fallback_targets or target_output_ports

        target_port = self.get_target_port(filtered_targets, routed_request)
        return (origin_port, target_port)

    def get_origin_port(
        self, possible_ports: list[port.Queue], routed_request: request.Request
    ) -> port.Queue:
        """
        Gets the most suitable origin port from the list of possible ports.

        Args:
            possible_ports (list[store.Queue]): List of possible origin ports.
            routed_request (request.Request): The routed request.

        Returns:
            store.Queue: The selected origin port.
        """
        # Distinguish primitives from products based on presence of product_info (primitives don't have it)
        if not hasattr(routed_request.requesting_item, 'product_info'):
            # Primitive: choose by heuristic
            origin_port = self.port_selection_heuristic(possible_ports, routed_request)
        else:
            # Product: find the concrete port that currently contains the item
            origin_port = get_port_with_item(
                possible_ports, routed_request.requesting_item.data.ID
            )
            if origin_port is None:
                raise ValueError(
                    f"No port found with item ID {routed_request.requesting_item.data.ID} in possible ports."
                )
        
        if origin_port is None:
            raise ValueError(
                f"No suitable port found for item ID {routed_request.requesting_item.data.ID} in possible ports."
            )
            
        if (
            origin_port.data.port_type == port_data.PortType.STORE
        ):  # Ports can have multiple input / output locations -> select the most suitable one
            if hasattr(origin_port, 'store_ports'):
                # This is a Store object with store_ports
                origin_port: port.Store
                origin_port = self.port_selection_heuristic(
                    origin_port.store_ports, routed_request
                )
            # If it's a StorePort, we don't need to do anything else
        return origin_port

    def get_target_port(
        self, possible_ports: list[port.Queue], routed_request: request.Request
    ) -> port.Queue:
        """
        Gets the most suitable target port from the list of possible ports.

        Args:
            possible_ports (list[store.Queue]): List of possible target ports.
            routed_request (request.Request): The routed request.

        Returns:
            store.Queue: The selected target port.
        """
        # TODO: consider here product_type compatibility logic
        target_port = self.port_selection_heuristic(possible_ports, routed_request)
        if target_port is None:
            raise ValueError(
                f"No suitable port found for item ID {routed_request.requesting_item.ID} in possible ports."
            )
        if target_port.data.port_type == port_data.PortType.STORE:
            if hasattr(target_port, 'store_ports'):
                # This is a Store object with store_ports
                target_port: port.Store
                target_port = self.port_selection_heuristic(
                    target_port.store_ports, routed_request
                )
            # If it's a StorePort, we don't need to do anything else
        return target_port


def get_port_with_item(ports: list[port.Queue], item_id: str) -> port.Queue:
    """
    Returns the port that contains the item with the given ID.

    Args:
        ports (list[store.Queue]): List of ports to search in.
        item_id (str): ID of the item to search for.

    Returns:
        store.Queue: The port that contains the item, or None if not found.
    """
    for port in ports:
        if item_id in port.items:
            return port
    return None


def random_port_selection_heuristic(
    possible_ports: list[port.Queue], routed_request: request.Request
) -> port.Queue:
    """
    Randomly selects a port from the given list of possible ports.

    Args:
        possible_ports (list[store.Queue]): List of possible ports to select from.

    Returns:
        store.Queue: A randomly selected port.
    """
    import random

    return random.choice(possible_ports)



from __future__ import annotations


import os
import time

from stable_baselines3 import PPO
from stable_baselines3.common.logger import configure
from stable_baselines3.common.callbacks import BaseCallback

import numpy as np

from gymnasium import spaces

from prodsys.adapters import ProductionSystemData
from prodsys.adapters import get_production_resources
from prodsys.control import sequencing_control_env
from prodsys.simulation import request


class ProductionControlEnv(sequencing_control_env.AbstractSequencingControlEnv):
    def get_observation(self) -> np.ndarray:
        """
        Function that utilizes the ResourceObserver of the environment class to get an array of observations of processes performed by the resource and in the queue of the resource. The observatino has a dimension CxP, where c is the capacity of resource and queue and P the number of processes.

        Returns:
            np.ndarray: The observation.
        """
        processes_observation = self.observer.observe_processes()
        encoded_processes = []
        processes = self.resource.data.process_ids

        for process_observation in processes_observation:
            encoded_process = [0 for _ in range(len(processes))]
            encoded_process[processes.index(process_observation.process)] = 1
            encoded_processes.append(encoded_process)

        encoded_process = [0 for _ in range(len(processes))]
        encoded_processes += [encoded_process] * (
            self.resource.data.capacity - len(processes_observation)
        )

        queue_observations = self.observer.observe_input_queue()
        for queue_observation in queue_observations:
            encoded_process = [0 for _ in range(len(processes))]
            encoded_process[processes.index(queue_observation.process)] = 1
            encoded_processes.append(encoded_process)

        encoded_process = [0 for _ in range(len(processes))]
        # v1 resources expose unified `ports` (queues); in this example we treat the first port as input.
        queue_capacity = self.resource.ports[0].capacity

        encoded_processes += [encoded_process] * (
            queue_capacity - len(queue_observations)
        )

        return np.array(encoded_processes)

    def get_info(self) -> dict:
        return {"info": 0}

    def get_termination_condition(self) -> bool:
        return self.runner.env.now >= 100000

    def get_reward(
        self, processed_request: request.Request, invalid_action: bool = False
    ) -> float:
        if invalid_action:
            reward = -1
        else:
            reward = (
                self.resource.current_setup is None
                or processed_request.process.process_data.ID
                == self.resource.current_setup.process_data.ID
            )
        if self.step_count % 10 == 0:
            reward += self.resource.ports[0].capacity - len(
                self.resource_controller.requests
            )

        return reward


class TensorboardCallback(BaseCallback):
    """
    Custom callback for plotting additional values in tensorboard.
    """

    def __init__(self, verbose=0):
        super(TensorboardCallback, self).__init__(verbose)

    def _on_step(self) -> bool:
        self.logger.record("reward", self.training_env.get_attr("reward")[0])
        self.logger.record("time", self.training_env.get_attr("runner")[0].env.now)
        return True


if __name__ == "__main__":
    adapter = ProductionSystemData.read(
        "examples/control/control_example_data/control_configuration.json"
    )
    # Pick a production resource robustly (first with any production process)
    production_resources = get_production_resources(adapter)
    if not production_resources:
        raise RuntimeError("No production resources found in configuration.")
    resource_data = production_resources[0]
    # Use the first declared port of the resource (validator ensures at least one)
    first_port_id = resource_data.ports[0]
    queue = [q for q in adapter.port_data if q.ID == first_port_id][0]
    shape = (int(queue.capacity) + resource_data.capacity, len(resource_data.process_ids))
    observation_space = spaces.Box(0, 1, shape=shape, dtype=int)
    action_space = spaces.Box(0, 1, shape=(int(queue.capacity),), dtype=float)
    env = ProductionControlEnv(
        adapter,
        resource_data.ID,
        observation_space=observation_space,
        action_space=action_space,
        render_mode="human",
    )

    tmp_path = os.path.join(
        os.getcwd(), "tensorboard_log", "sequencing", time.strftime("%Y%m%d-%H%M%S")
    )
    new_logger = configure(tmp_path, ["stdout", "csv", "tensorboard"])

    model = PPO(env=env, policy="MlpPolicy", verbose=1)
    model.set_logger(new_logger)
    model.learn(total_timesteps=1000000, callback=TensorboardCallback())

    # Start Tensorboard with: tensorboard --logdir tensorboard_log\sequncing



import pytest
from prodsys.models.production_system_data import ProductionSystemData
import prodsys.express as psx
from prodsys import runner


@pytest.fixture
def simulation_adapter() -> ProductionSystemData:
    t1 = psx.FunctionTimeModel("exponential", 0.8, 0, "t1")
    t2 = psx.FunctionTimeModel("exponential", 1.2, 0, "t2")

    p1 = psx.ProductionProcess(t1, "p1")
    p2 = psx.ProductionProcess(t2, "p2")

    # t3 = psx.FunctionTimeModel("constant", 0.1, ID="t3")
    t3 = psx.DistanceTimeModel(60, 0.05, "manhattan", ID="t3")

    tp = psx.TransportProcess(t3, "tp")
    tp_primitive = psx.TransportProcess(t3, "tp_primitive")

    machine = psx.Resource([p1], [5, 0], 1, ID="machine")
    machine2 = psx.Resource([p2], [5, 5], 2, ID="machine2")

    transport = psx.Resource([tp], [3, 0], 1, ID="transport")
    transport2 = psx.Resource([tp], [3, 0], 1, ID="transport2")
    transport_primitive = psx.Resource([tp_primitive], [4, 0], 1, ID="transport_primitive")

    storage1 = psx.Store(ID="storage1", location=[6, 0], capacity=30)
    storage2 = psx.Store(ID="storage2", location=[11, 0], capacity=20)

    primitive1 = psx.Primitive(
        ID="primitive1",
        transport_process=tp_primitive,
        storages=[storage1],
        quantity_in_storages=[10],
    )

    primitive2 = psx.Primitive(
        ID="primitive2",
        transport_process=tp_primitive,
        storages=[storage2],
        quantity_in_storages=[20],
    )

    primitive_dependency_1 = psx.ToolDependency(
        ID="primitive_dependency_1",
        required_entity=primitive1,
    )
    primitive_dependency_2 = psx.ToolDependency(
        ID="primitive_dependency_2",
        required_entity=primitive2,
    )

    product1 = psx.Product(
        process=[p1, p2],
        transport_process=tp,
        ID="product1",
        dependencies=[primitive_dependency_1],
    )
    product2 = psx.Product(
        process=[p2, p1],
        transport_process=tp,
        ID="product2",
        dependencies=[primitive_dependency_2],
    )

    sink1 = psx.Sink(product1, [10, 0], "sink1")
    sink2 = psx.Sink(product2, [10, 0], "sink2")

    arrival_model_1 = psx.FunctionTimeModel("constant", 2.6, ID="arrival_model_1")
    arrival_model_2 = psx.FunctionTimeModel("constant", 1.3, ID="arrival_model_2")

    source1 = psx.Source(product1, arrival_model_1, [0, 0], ID="source_1")
    source2 = psx.Source(product2, arrival_model_2, [0, 0], ID="source_2")

    system = psx.ProductionSystem(
        [machine, machine2, transport, transport2, transport_primitive],
        [source1, source2],
        [sink1, sink2],
        [primitive1, primitive2],
    )
    system.validate()
    adapter = system.to_model()
    return adapter


def test_initialize_simulation(simulation_adapter: ProductionSystemData):
    runner_instance = runner.Runner(production_system_data=simulation_adapter)
    runner_instance.initialize_simulation()


def test_hashing(simulation_adapter: ProductionSystemData):
    hash_str = simulation_adapter.hash()
    assert hash_str == "dbd3cb1cd2ceff29d27b40af9216cd91"


def test_run_simulation(simulation_adapter: ProductionSystemData):
    runner_instance = runner.Runner(production_system_data=simulation_adapter)
    runner_instance.initialize_simulation()
    runner_instance.run(1000)
    assert runner_instance.env.now == 1000
    runner_instance.print_results()
    post_processor = runner_instance.get_post_processor()
    for kpi in post_processor.throughput_and_output_KPIs:
        if kpi.name == "output" and kpi.product_type == "product1":
            assert kpi.value > 370 and kpi.value < 390
        if kpi.name == "output" and kpi.product_type == "product2":
            assert kpi.value > 740 and kpi.value < 770
    for kpi in post_processor.machine_state_KPIS:
        if kpi.name == "productive_time" and kpi.resource == "machine":
            assert kpi.value < 94 and kpi.value > 91

        if kpi.name == "productive_time" and kpi.resource == "transport":
            assert kpi.value > 48 and kpi.value < 55
        if kpi.name == "productive_time" and kpi.resource == "transport2":
            assert kpi.value > 48 and kpi.value < 55
        if kpi.name == "productive_time" and kpi.resource == "transport_primitive":
            assert kpi.value > 73 and kpi.value < 77

    for kpi in post_processor.WIP_KPIs:
        if kpi.name == "WIP" and kpi.product_type == "product1":
            assert kpi.value < 5.5 and kpi.value > 4.0
        if kpi.name == "WIP" and kpi.product_type == "product2":
            assert kpi.value < 10 and kpi.value > 8

    for kpi in post_processor.primitive_WIP_KPIs:
        if kpi.name == "primitive_WIP" and kpi.product_type == "primitive1":
            assert kpi.value < 5.5 and kpi.value > 4.0
        if kpi.name == "primitive_WIP" and kpi.product_type == "primitive2":
            assert kpi.value < 10 and kpi.value > 8

    for kpi in post_processor.aggregated_throughput_time_KPIs:
        if kpi.name == "throughput_time" and kpi.product_type == "product1":
            assert kpi.value < 12.5 and kpi.value > 10.5
        if kpi.name == "throughput_time" and kpi.product_type == "product2":
            assert kpi.value < 12.5 and kpi.value > 10.5



from typing import Dict, List, Optional, Union, Set


from pydantic import Field, PrivateAttr
from pydantic.dataclasses import dataclass

from prodsys.express.primitive import Primitive
from prodsys.util import util
from prodsys.models import production_system_data

from prodsys.express import (
    core,
    node,
    resources,
    source,
    time_model,
    state,
    process,
    product,
    sink,
)


def remove_duplicate_items(
    items: List[
        Union[
            resources.Resource,
            source.Source,
            sink.Sink,
            product.Product,
            state.STATE_UNION,
            process.PROCESS_UNION,
            time_model.TIME_MODEL_UNION,
        ]
    ],
) -> List[
    Union[
        resources.Resource,
        source.Source,
        sink.Sink,
        product.Product,
        state.STATE_UNION,
        process.PROCESS_UNION,
        time_model.TIME_MODEL_UNION,
    ]
]:
    """
    Removes duplicate items from a list of items.
    """
    id_set = set()
    filtered_items = []
    for item in items:
        if item.ID in id_set:
            continue
        id_set.add(item.ID)
        filtered_items.append(item)
    return filtered_items


@dataclass
class ProductionSystem(core.ExpressObject):
    """
    Class that represents a production system. A production system containts of resources, products, sources and sinks.
    It is the `prodsys.express` equivalent to the 'ProductionSystemAdapter' of the `prodsys.adapters` module and
    can be converted to this data object. In contrast to the adapter,
    this class nests the objects in a tree structure, which makes it easier to work with when instantiating
    a production system, but more complicated when reviewing the data itself.

    Args:
        resources (List[resources.Resource]): Resources of the production system.
        sources (List[source.Source]): Sources of the production system.
        sinks (List[sink.Sink]): Sinks of the production system.
    """

    resources: List[Union[resources.Resource]]
    sources: List[Union[source.Source, source.OrderSource]]
    sinks: List[sink.Sink]
    primitives: List[Primitive] = Field(default_factory=list)

    def __post_init__(self):
        self._runner: Optional[runner.Runner] = None

    def to_model(self) -> production_system_data.ProductionSystemData:
        """
        Converts the `prodsys.express` object to a data object from `prodsys.models`.

        Returns:
            prodsys.adapters.Adapter: An instance of the data object.
        """
        # Collect products from regular sources (OrderSource doesn't have a product attribute)
        from prodsys.express.source import Source, OrderSource
        products = [
            src.product for src in self.sources 
            if isinstance(src, Source) and hasattr(src, 'product')
        ] + [
            sink.product for sink in self.sinks
        ]
        # Collect products from OrderSource orders
        for src in self.sources:
            if isinstance(src, OrderSource):
                for order in src.orders:
                    for ordered_product in order.ordered_products:
                        products.append(ordered_product.product)
        products = remove_duplicate_items(products)
        dependencies = list(
            util.flatten_object([product.dependencies for product in products])
        )
        dependencies += list(
            util.flatten_object([resource.dependencies for resource in self.resources])
        )
        dependencies = remove_duplicate_items(dependencies)
        process_candidates = list(
            util.flatten_object(
                [product.process for product in products]
                + [product.transport_process for product in products]
                + [primitive.transport_process for primitive in self.primitives]
                + [resource.processes for resource in self.resources]
            )
        )
        
        processes: List[process.PROCESS_UNION] = []
        visited_process_ids: Set[str] = set()

        def add_process_instance(process_instance: process.PROCESS_UNION) -> None:
            if process_instance is None or not hasattr(process_instance, "ID"):
                return
            process_id = process_instance.ID
            if process_id in visited_process_ids:
                return
            visited_process_ids.add(process_id)
            processes.append(process_instance)
            if isinstance(process_instance, process.ProcessModel):
                for nested_process in process_instance.get_all_nested_processes():
                    add_process_instance(nested_process)

        for process_candidate in process_candidates:
            add_process_instance(process_candidate)

        # Collect processes from states (SetupState, ProcessBreakdownState, etc.)
        states = list(
            util.flatten_object([resource.states for resource in self.resources])
        )
        states = remove_duplicate_items(states)
        
        # Extract processes from states
        state_processes = []
        for state_instance in states:
            if isinstance(state_instance, state.SetupState):
                state_processes.append(state_instance.origin_setup)
                state_processes.append(state_instance.target_setup)
            elif isinstance(state_instance, state.ProcessBreakdownState):
                state_processes.append(state_instance.process)
        
        # Add processes from states to the processes list
        for state_process_instance in state_processes:
            add_process_instance(state_process_instance)
        
        # Build a mapping of process IDs to process objects for resolving ProcessModel references
        process_id_to_process: Dict[str, process.PROCESS_UNION] = {
            proc.ID: proc for proc in processes
        }

        # Extract processes referenced in ProcessModel adjacency matrices
        process_model_processes = []
        for proc in processes:
            if isinstance(proc, process.ProcessModel):
                # Extract all process IDs from the adjacency matrix
                process_ids = set()
                for key in proc.adjacency_matrix.keys():
                    process_ids.add(key)
                for value_list in proc.adjacency_matrix.values():
                    process_ids.update(value_list)
                
                # Find process objects by ID from our mapping
                for process_id in process_ids:
                    if process_id in process_id_to_process:
                        # Already in the list, skip
                        continue
                    # Try to find the process in resources' processes
                    for resource_instance in self.resources:
                        for resource_process in resource_instance.processes:
                            if resource_process.ID == process_id and resource_process not in processes:
                                process_id_to_process[process_id] = resource_process
                                process_model_processes.append(resource_process)
                                break
                        # Also check states for this process
                        for state_instance in resource_instance.states:
                            if isinstance(state_instance, state.SetupState):
                                if state_instance.origin_setup.ID == process_id and state_instance.origin_setup not in processes:
                                    process_id_to_process[process_id] = state_instance.origin_setup
                                    process_model_processes.append(state_instance.origin_setup)
                                if state_instance.target_setup.ID == process_id and state_instance.target_setup not in processes:
                                    process_id_to_process[process_id] = state_instance.target_setup
                                    process_model_processes.append(state_instance.target_setup)
                            elif isinstance(state_instance, state.ProcessBreakdownState):
                                if state_instance.process.ID == process_id and state_instance.process not in processes:
                                    process_id_to_process[process_id] = state_instance.process
                                    process_model_processes.append(state_instance.process)
        
        # Add processes found from ProcessModel adjacency matrices
        for process_model_process in process_model_processes:
            add_process_instance(process_model_process)
        processes = remove_duplicate_items(processes)
        process_id_to_process = {proc.ID: proc for proc in processes}
        dependencies += list(
            util.flatten_object(
                [
                    process_instance.dependencies
                    for process_instance in processes
                    if not isinstance(
                        process_instance, process.RequiredCapabilityProcess
                    )
                ]
            )
        )

        product_data_models = [product.to_model() for product in products]

        required_process_ids_for_resources: Set[str] = set()
        required_capabilities: Set[str] = set()

        def include_process_id(process_id: Optional[str]) -> None:
            if not process_id:
                return
            process_obj = process_id_to_process.get(process_id)
            if isinstance(process_obj, process.RequiredCapabilityProcess):
                capability = getattr(process_obj, "capability", None)
                if capability:
                    required_capabilities.add(capability)
                return
            required_process_ids_for_resources.add(process_id)
            capability = getattr(process_obj, "capability", None)
            if capability:
                required_capabilities.add(capability)

        for product_data_model in product_data_models:
            processes_field = product_data_model.processes
            if processes_field is None:
                continue
            if isinstance(processes_field, dict):
                for node_id, successors in processes_field.items():
                    include_process_id(node_id)
                    if successors:
                        for successor in successors:
                            include_process_id(successor)
            else:
                flattened = util.flatten_object(processes_field)
                if flattened:
                    for process_id in flattened:
                        include_process_id(process_id)
            include_process_id(product_data_model.transport_process)

        for state_process_instance in state_processes:
            if state_process_instance is None or not hasattr(
                state_process_instance, "ID"
            ):
                continue
            if isinstance(state_process_instance, process.CapabilityProcess):
                if state_process_instance.capability:
                    required_capabilities.add(state_process_instance.capability)
            else:
                capability = getattr(state_process_instance, "capability", None)
                if capability:
                    required_capabilities.add(capability)

        nodes = []
        for process_instance in processes:
            if isinstance(process_instance, process.RequiredCapabilityProcess):
                continue
            for dependency in process_instance.dependencies:
                if not hasattr(dependency, "interaction_node"):
                    continue
                if dependency.interaction_node:
                    nodes.append(dependency.interaction_node)
            if not isinstance(process_instance, process.LinkTransportProcess):
                continue
            for link in process_instance.links:
                for link_element in link:
                    if isinstance(link_element, node.Node):
                        nodes.append(link_element)
        for resource_instance in self.resources:
            for dependency in resource_instance.dependencies:
                if not hasattr(dependency, "interaction_node"):
                    continue
                if dependency.interaction_node:
                    nodes.append(dependency.interaction_node)
        nodes = remove_duplicate_items(nodes)
        primitive_stores = list(
            util.flatten_object(
                primitive_store
                for primitive_store in [
                    primitive.storages for primitive in self.primitives
                ]
            )
        )
        primitive_stores = remove_duplicate_items(primitive_stores)
        # States are already collected above, no need to collect again

        time_models = (
            [
                process_instance.time_model
                for process_instance in processes
                if not isinstance(process_instance, (process.RequiredCapabilityProcess, process.ProcessModel))
            ]
            + [state_instance.time_model for state_instance in states]
            + [src.time_model for src in self.sources if isinstance(src, Source) and hasattr(src, 'time_model')]
        )
        time_models += [
            process_instance.loading_time_model
            for process_instance in processes
            if isinstance(process_instance, process.TransportProcess)
            and process_instance.loading_time_model
        ]
        time_models += [
            process_instance.unloading_time_model
            for process_instance in processes
            if isinstance(process_instance, process.TransportProcess)
            and process_instance.unloading_time_model
        ]
        time_models += [
            s.repair_time_model
            for s in states
            if isinstance(s, state.BreakDownState)
            or isinstance(s, state.ProcessBreakdownState)
        ]
        time_models += [
            s.battery_time_model for s in states if isinstance(s, state.ChargingState)
        ]
        time_models += [
            s.non_scheduled_time_model for s in states if isinstance(s, state.NonScheduledState)
        ]

        time_models = remove_duplicate_items(time_models)

        time_model_data = [time_model.to_model() for time_model in time_models]
        process_data = [process.to_model() for process in processes]
        state_data = [state.to_model() for state in states]
        nodes_data = [node.to_model() for node in nodes]
        product_data = product_data_models
        resource_model_data = []
        provided_process_ids: Set[str] = set()
        provided_capabilities: Set[str] = set()

        for resource_instance in self.resources:
            resource_model = resource_instance.to_model()
            resource_model_data.append(resource_model)
            for proc_instance in resource_instance.processes:
                if getattr(proc_instance, "ID", None):
                    provided_process_ids.add(proc_instance.ID)
                capability = getattr(proc_instance, "capability", None)
                if capability:
                    provided_capabilities.add(capability)

        missing_process_ids = required_process_ids_for_resources - provided_process_ids
        missing_capabilities = required_capabilities - provided_capabilities

        if missing_process_ids or missing_capabilities:
            message_fragments = []
            if missing_process_ids:
                message_fragments.append(
                    f"Missing resources for processes: {sorted(missing_process_ids)}"
                )
            if missing_capabilities:
                message_fragments.append(
                    f"Missing resources providing capabilities: {sorted(missing_capabilities)}"
                )
            raise ValueError(
                "Production system configuration invalid after process assignment. "
                + " ".join(message_fragments)
            )

        resource_data = resource_model_data
        source_data = [source.to_model() for source in self.sources]
        sink_data = [sink.to_model() for sink in self.sinks]
        
        # Collect orders from OrderSource
        from prodsys.express.source import OrderSource
        order_data_list = []
        for src in self.sources:
            if isinstance(src, OrderSource):
                # Use the stored order models from to_model()
                if hasattr(src, '_order_models'):
                    order_data_list.extend(src._order_models)
                else:
                    # Fallback: convert orders to models
                    order_data_list.extend([order.to_model() for order in src.orders])
        # Remove duplicate orders by ID
        if order_data_list:
            seen_order_ids = set()
            unique_orders = []
            for order in order_data_list:
                if order.ID not in seen_order_ids:
                    seen_order_ids.add(order.ID)
                    unique_orders.append(order)
            order_data_list = unique_orders if unique_orders else None
        else:
            order_data_list = None
        dependencies = remove_duplicate_items(dependencies)
        dependency_data = [dependency.to_model() for dependency in dependencies]
        primitive_data = [
            primitive.to_model() for primitive in self.primitives if primitive.storages
        ]
        primitive_storage_data = [storage.to_model() for storage in primitive_stores]

        ports = list(
            util.flatten_object(
                [s.ports for s in self.sources]
                + [r.ports for r in self.resources if r.ports]
                + [r.buffers for r in self.resources if r.buffers]
                + [s.ports for s in self.sinks]
            )
        )
        port_data = [q.to_model() for q in ports]
        port_data = remove_duplicate_items(port_data)
        return production_system_data.ProductionSystemData(
            time_model_data=time_model_data,
            process_data=process_data,
            state_data=state_data,
            node_data=nodes_data,
            product_data=product_data,
            nodes_data=nodes_data,
            resource_data=resource_data,
            source_data=source_data,
            sink_data=sink_data,
            port_data=port_data + primitive_storage_data,
            dependency_data=dependency_data,
            primitive_data=primitive_data,
            order_data=order_data_list,
        )

    def run(self, time_range: float = 2880, seed: int = 0):
        """
        Runs the simulation of the production system.

        Args:
            time_range (float, optional): The time range of the simulation. Defaults to 2880.
            seed (int, optional): The seed of the simulation. Defaults to 0.
        """
        self._runner = runner.Runner(production_system_data=self.to_model())
        self._runner.production_system_data.seed = seed
        self._runner.initialize_simulation()
        self._runner.run(time_range)

    def validate(self):
        """
        Validates the production system. Checks if the production system is valid.

        Raises:
            ValueError: If the production system is not valid.
        """
        adapter = self.to_model()
        adapter.validate_configuration()

    @property
    def runner(self):
        if not self._runner:
            raise ValueError(
                "Runner has not been initialized. Please run the simulation first with the run function."
            )
        return self._runner

    @property
    def post_processor(self):
        if not self._runner:
            raise ValueError(
                "Runner has not been initialized. Please run the simulation first with the run function."
            )
        return self._runner.get_post_processor()


from prodsys.simulation import runner
